import logging

import pytest

try:
    import django
    from django.conf import settings
    from django.test import override_settings
except ImportError:
    django = None

from parasolr import django
from parasolr.schema import SolrSchema


logger = logging.getLogger(__name__)


if django:


    def get_test_solr_config():
        '''Get configuration for test Solr connection based on
        default and test options in django settings. Any test configuration
        options specified are used; if no test collection name
        is specified, generates one based on the configured collection.'''

        # copy default config for basic connection optiosn (e.g. url)
        test_config = settings.SOLR_CONNECTIONS['default'].copy()

        # use test settings as primary: anything in test settings
        # should override default settings
        if 'TEST' in settings.SOLR_CONNECTIONS['default']:
            test_config.update(settings.SOLR_CONNECTIONS['default']['TEST'])

        # if test collection is not explicitly configured,
        # set it based on default collection
        if 'TEST' not in test_config or \
         'COLLECTION' not in settings.SOLR_CONNECTIONS['default']['TEST']:
            test_config['COLLECTION'] = 'test_%s' % \
                settings.SOLR_CONNECTIONS['default']['COLLECTION']

        logger.info('Configuring Solr for tests %(URL)s%(COLLECTION)s',
                    test_config)
        return test_config

    @pytest.fixture(autouse=True, scope="session")
    def configure_django_test_solr():
        """Automatically configure the default Solr to use a test
        core based on the configured **SOLR_CONNECTIONS**.  Will use
        test name if specified (using the same structure as Django
        DATABASES), or prepend "test_" to the configured COLLECTION
        if no test name is set. The test core will be created and
        schema configured before starting, and unloaded after tests
        complete.  Example configuration::

            SOLR_CONNECTIONS = {
                'default': {
                    'URL': 'http://localhost:8983/solr/',
                    'COLLECTION': 'myproj',
                    'TEST': {
                        'NAME': 'testproj',
                        }
                }
            }

        """

        solr_config_opts = get_test_solr_config()

        logger.info('Configuring Solr for tests %(URL)s%(COLLECTION)s',
                    solr_config_opts)

        with override_settings(SOLR_CONNECTIONS={'default': solr_config_opts}):
            # reload core before and after to ensure field list is accurate
            solr = django.SolrClient(commitWithin=10)
            response = solr.core_admin.status(core=solr_config_opts['COLLECTION'])
            if not response.status.get(solr_config_opts['COLLECTION'], None):
                solr.core_admin.create(solr_config_opts['COLLECTION'],
                                       configSet=solr_config_opts.get('CONFIGSET', 'basic_configs'))

            try:
                # if a schema is configured, update the test core
                schema_config = SolrSchema.get_configuration()
                schema_config.configure_fieldtypes(solr)
                schema_config.configure_fields(solr)
            except Exception:
                pass

            # yield settings so tests run with overridden solr connection
            yield settings

            # clear out any data indexed in test collection
            solr.update.delete_by_query('*:*')
            # and unload
            solr.core_admin.unload(
                solr_config_opts['COLLECTION'],
                deleteInstanceDir=True,
                deleteIndex=True,
                deleteDataDir=True
            )



"""
Object-oriented approach to Solr searching and filtering modeled
on :class:`django.models.queryset.QuerySet`.  Supports iteration,
slicing, counting, and boolean check to see if a search has results.

Filter, search and sort methods return a new queryset, and can be
chained. For example::

    SolrQuerySet(solrclient).filter(item_type='person') \
                            .search(name='hem*') \
                            .order_by('sort_name') \


If you are working with Django you should use
:class:`parasolr.django.SolrQuerySet`,
which will automatically initialize a new :class:`parasolr.django.SolrClient`
if one is not passed in.
"""
from collections import OrderedDict
from typing import Dict, List

from parasolr.solr import SolrClient
from parasolr.solr.client import QueryResponse

class SolrQuerySet:
    """A Solr queryset object that allows for object oriented
    searching and filtering of Solr results. Allows search results
    to be pagination using slicing, count, and iteration.

    """

    _result_cache = None
    start = 0
    stop = None
    sort_options = []
    search_qs = []
    filter_qs = []
    field_list = []
    highlight_field = None
    facet_field = []
    facet_opts = {}
    highlight_opts = {}
    raw_params = {}


    #: by default, combine search queries with AND
    default_search_operator = 'AND'

    def __init__(self, solr: SolrClient):
        # requires solr client so that this version can be django-agnostic
        self.solr = solr
        # convert search operator into form needed for combining queries
        self._search_op = ' %s ' % self.default_search_operator

    def get_results(self, **kwargs) -> List[dict]:
        """
        Query Solr and get the results for the current query and filter
        options. Populates result cache and returns the documents portion
        of the reponse.

        Returns:
            Solr response documents as a list of dictionaries.
        """

        # TODO: can we store the result cache and only retrieve
        # if query options have changed?
        # For now, always query.

        query_opts = self.query_opts()
        query_opts.update(**kwargs)
        # TODO: what do we do about the fact that Solr defaults
        # to 10 rows?

        # NOTE: django templates choke on AttrDict because it is
        # callable; using dictionary response instead
        self._result_cache = self.solr.query(**query_opts)
        return [doc.as_dict() for doc in self._result_cache.docs]

    def query_opts(self) -> Dict[str, str]:
        """Construct query options based on current queryset configuration.
        Includes filter queries, start and rows, sort, and search query.
        """
        query_opts = {
            'start': self.start,
        }

        if self.filter_qs:
            query_opts['fq'] = self.filter_qs
        if self.stop:
            query_opts['rows'] = self.stop - self.start
        if self.sort_options:
            query_opts['sort'] = ','.join(self.sort_options)

        # main query; if no query is defined, find everything
        if self.search_qs:
            query_opts['q'] = self._search_op.join(self.search_qs)
        else:
            query_opts['q'] = '*:*'

        if self.field_list:
            query_opts['fl'] = ','.join(self.field_list)

        if self.highlight_field:
            query_opts.update({
                'hl': True,
                'hl.field': self.highlight_field
            })
            for key, val in self.highlight_opts.items():
                query_opts['hl.%s' % key] = val

        if self.facet_field:
            query_opts.update({
                'facet': True,
                'facet.field': self.facet_field
            })
            for key, val in self.facet_opts.items():
                query_opts['facet.%s' % key] = val

        # include any raw query parameters
        query_opts.update(self.raw_params)

        return query_opts

    def count(self) -> int:
        """Total number of results for the current query"""

        # if result cache is already populated, use it
        if self._result_cache is not None:
            return self._result_cache.numFound

        # otherwise, query with current options but request zero rows
        # and do not populate the result cache
        query_opts = self.query_opts()
        # setting these by dictionary assignment, because conflicting
        # kwargs results in a Python exception
        query_opts['rows'] = 0
        query_opts['facet'] = False
        query_opts['hl'] = False
        return self.solr.query(**query_opts).numFound

    def get_facets(self) -> Dict[str, int]:
        """Return a dictionary of facets and their values and
        counts as key/value pairs.
        """
        if self._result_cache is not None:
            # wrap to process facets and return as dictionary
            # for Django template support
            qr = QueryResponse(self._result_cache)
            # NOTE: using dictionary syntax preserves OrderedDict
            return qr.facet_counts['facet_fields']
        # since we just want a dictionary of facet fields, don't populate
        # the result cache, no rows needed

        query_opts = self.query_opts()
        query_opts['rows'] = 0
        query_opts['hl'] = False
        # setting these by dictionary assignment, because conflicting
        # kwargs results in a Python exception
        return self.solr.query(**query_opts).facet_counts['facet_fields']

    @staticmethod
    def _lookup_to_filter(key, value) -> str:
        """Convert keyword argument key=value pair into a Solr filter.
        Currently only supports simple case of field:value."""

        # NOTE: as needed, we can start implementing django-style filters
        # such as __in=[a, b, c] or __range=(start, end)
        return '%s:%s' % (key, value)

    def filter(self, *args, **kwargs) -> 'SolrQuerySet':
        """
        Return a new SolrQuerySet with Solr filter queries added.
        Multiple filters can be combined either in a single
        method call, or they can be chained for the same effect.
        For example::

            queryset.filter(item_type='person').filter(birth_year=1900)
            queryset.filter(item_type='person', birth_year=1900)

        To provide a filter that should be used in modified, provide
        the exact string of your filter query::

            queryset.filter('birth_year:[1800 TO *]')

        """
        qs_copy = self._clone()

        # any args are treated as filter queries without modification
        qs_copy.filter_qs.extend(args)

        for key, value in kwargs.items():
            qs_copy.filter_qs.append(self._lookup_to_filter(key, value))

        return qs_copy

    def facet(self, *args: str, **kwargs) -> 'SolrQuerySet':
        """
        Request facets for specified fields. Returns a new SolrQuerySet
        with Solr faceting enabled and facet.field parameter set. Does not
        support ranged faceting.

        Subsequent calls will reset the facet.field to the last set of
        args in the chain.

        For example::

            qs = queryset.facet('person_type', 'age')
            qs = qs.facet('item_type')

        would result in `item_type` being the only facet field.
        """
        qs_copy = self._clone()

        # cast args tuple to list for consistency with other iterable fields
        qs_copy.facet_field = list(args)
        # add other kwargs to be prefixed in query_opts
        qs_copy.facet_opts.update(kwargs)

        return qs_copy

    def search(self, *args, **kwargs) -> 'SolrQuerySet':
        """
        Return a new SolrQuerySet with search queries added. All
        queries will combined with the default search operator when
        constructing the `q` parameter sent to Solr..
        """
        qs_copy = self._clone()
        # any args are treated as search queries without modification
        qs_copy.search_qs.extend(args)

        for key, value in kwargs.items():
            qs_copy.search_qs.append(self._lookup_to_filter(key, value))

        return qs_copy

    def order_by(self, *args) -> 'SolrQuerySet':
        """Apply sort options to the queryset by field name. If the field
        name starts with -, sort is descending; otherwise ascending."""
        qs_copy = self._clone()
        for sort_option in args:
            if sort_option.startswith('-'):
                sort_order = 'desc'
                sort_option = sort_option.lstrip('-')
            else:
                sort_order = 'asc'
            qs_copy.sort_options.append('%s %s' % (sort_option, sort_order))

        return qs_copy

    def query(self, **kwargs) -> 'SolrQuerySet':
        """Return a new SolrQuerySet with the results populated from Solr.
        Any options passed in via keyword arguments take precedence
        over query options on the queryset.
        """
        qs_copy = self._clone()
        qs_copy.get_results(**kwargs)
        return qs_copy

    def only(self, *args, **kwargs) -> 'SolrQuerySet':
        """Use field limit option to return only the specified fields.
        Optionally provide aliases for them in the return. Subsequent
        calls will *replace* any previous field limits. Example::

            queryset.only('title', 'author', 'date')
            queryset.only('title:title_t', 'date:pubyear_i')

        """
        qs_copy = self._clone()
        # *replace* any existing field list with the current values
        qs_copy.field_list = list(args)
        for key, value in kwargs.items():
            qs_copy.field_list.append('%s:%s' % (key, value))

        return qs_copy

    def highlight(self, field: str, **kwargs) -> 'SolrQuerySet':
        """"Configure highlighting. Takes arbitrary Solr highlight
        parameters and adds the `hl.` prefix to them.  Example use::

            queryset.highlight('content', snippets=3, method='unified')
        """
        qs_copy = self._clone()
        qs_copy.highlight_field = field
        qs_copy.highlight_opts = kwargs
        return qs_copy

    def raw_query_parameters(self, **kwargs) -> 'SolrQuerySet':
        """Add abritrary raw parameters to be included in the query
        request, e.g. for variables referenced in join or field queries.
        Analogous to the input of the same name in the Solr web interface."""
        qs_copy = self._clone()
        qs_copy.raw_params.update(kwargs)
        return qs_copy

    def get_highlighting(self):
        """Return the highlighting portion of the Solr response."""
        if not self._result_cache:
            self.get_results()
        return self._result_cache.get('highlighting', {})

    def all(self) -> 'SolrQuerySet':
        """Return a new queryset that is a copy of the current one."""
        return self._clone()

    def none(self) -> 'SolrQuerySet':
        """Return an empty result list."""
        qs_copy = self._clone()
        # replace any search queries with this to find not anything
        qs_copy.search_qs = ['NOT *:*']
        return qs_copy

    def _clone(self) -> 'SolrQuerySet':
        """
        Return a copy of the current QuerySet for modification via
        filters.
        """
        # create a new instance with same solr and query opts
        # use current class to support extending
        qs_copy = self.__class__(solr=self.solr)
        # set attributes that can be copied directly
        qs_copy.start = self.start
        qs_copy.stop = self.stop
        qs_copy.highlight_field = self.highlight_field

        # set copies of list and dict attributes
        qs_copy.search_qs = list(self.search_qs)
        qs_copy.filter_qs = list(self.filter_qs)
        qs_copy.sort_options = list(self.sort_options)
        qs_copy.field_list = list(self.field_list)
        qs_copy.highlight_opts = dict(self.highlight_opts)
        qs_copy.raw_params = dict(self.raw_params)
        qs_copy.facet_field = list(self.facet_field)
        qs_copy.facet_opts = dict(self.facet_opts)


        return qs_copy

    def set_limits(self, start, stop):
        """Set limits to get a subsection of the results, to support slicing."""
        if start is None:
            start = 0
        self.start = start
        self.stop = stop

    iter_chunk_size = 1000

    def __iter__(self):
        """Iterate over result documents for this query."""
        return iter(self.get_results())

    def __bool__(self):
        """results are not empty"""
        return bool(self.get_results())

    def __getitem__(self, k):
        """Return a single result or a slice of results"""
        # based on django queryset logic

        if not isinstance(k, (int, slice)):
            raise TypeError
        assert ((not isinstance(k, slice) and (k >= 0)) or
                (isinstance(k, slice) and (k.start is None or k.start >= 0) and
                 (k.stop is None or k.stop >= 0))), \
            "Negative indexing is not supported."

        # if the result cache is already populated,
        # return the requested index or slice
        if self._result_cache is not None:
            return self._result_cache.docs[k]

        qs_copy = self._clone()

        if isinstance(k, slice):
            if k.start is not None:
                start = int(k.start)
            else:
                start = None
            if k.stop is not None:
                stop = int(k.stop)
            else:
                stop = None

            qs_copy.set_limits(start, stop)
            return list(qs_copy)[::k.step] if k.step else qs_copy

        # single item
        qs_copy.set_limits(k, k + 1)
        return qs_copy.get_results()[0]



"""
Module with class and methods for the Solr Schema API.
"""
from typing import Any, Optional, List
from urllib.parse import urljoin

from attrdict import AttrDict
import requests

from parasolr.solr.client import ClientBase

class Schema(ClientBase):
    """Class for managing Solr Schema API

    Args:
        solr_url: Base url for Solr.
        collection: Name of the collection or core.
        handler: Handler name for Solr Schema API.
        session: A python-requests :class:`requests.Session`.
    """
    def __init__(self, solr_url: str, collection: str, handler: str,
                 session: requests.Session=None):
        # Go ahead and create a session if one is not passed in
        super().__init__(session=session)
        self.url = self.build_url(solr_url, collection, handler)
        self.headers = {
            'Content-type': 'application/json'
        }

    def _post_field(self, method: str, **field_kwargs: Any) -> None:
        """Post a field definition to the schema API.

        Args:
          method: Solr field method to use.
          **field_kwargs: Field arguments to use in definition. Any valid schema
            definition may be used; if passed as ``kwargs``, rather than
            :class:`dict`, ``klass`` may be used instead of ``class``.
        """
        # Handle situations where we need class as a kwarg
        if 'klass' in field_kwargs:
            field_kwargs['class'] = field_kwargs['klass']
            del field_kwargs['klass']
        data = {
            method: field_kwargs
        }
        self.make_request(
            'post',
            self.url,
            headers=self.headers,
            data=data
        )

    def add_field(self, **field_kwargs: Any) -> None:
        """Add a field with the supplied definition.

        Args:
          **field_kwargs: Any valid Solr field definition values.
        """
        self._post_field('add-field', **field_kwargs)

    def delete_field(self, name: str) -> None:
        """Delete a field with the supplied name.

        Args:
          name: Name of field to delete.
        """
        self._post_field('delete-field', name=name)

    def replace_field(self, **field_kwargs: Any) -> None:
        """Replace a field with the supplied definition

        Args:
          **field_kwargs: Any valid Solr field definition values; must be a
            full redefinition, not a partial update.
        """
        self._post_field('replace-field', **field_kwargs)

    def add_copy_field(self, source: str, dest: str,
                       maxChars: int=None) -> None:
        """Add a copy field between two existing fields.

        Args:
            source: Source Solr field.
            dest: Destination Solr field.
            maxChars: Maximum characters to copy.
        """
        field_definition = {
            'source': source,
            'dest': dest
        }
        if maxChars:
            field_definition['maxChars'] = maxChars
        self._post_field('add-copy-field', **field_definition)

    def delete_copy_field(self, source: str, dest: str) -> None:
        """Delete a Solr copy field.

        Args:
            source: Source Solr field.
            dest: Destination Solr field.
        """
        self._post_field(
            'delete-copy-field',
            **{'source': source, 'dest': dest}
        )

    def add_field_type(self, **field_kwargs: Any) -> None:
        """Add a field type to a Solr collection or core.

        Args:
            **field_kwargs: Any valid Solr field definition values.

        Returns:
            None
        """
        self._post_field('add-field-type', **field_kwargs)

    def delete_field_type(self, name: str) -> None:
        """Delete a field type from a Solr collection or core.

        Args:
            name: Name of Solr field type to delete.
        """
        self._post_field('delete-field-type', name=name)

    def replace_field_type(self, **field_kwargs: Any) -> None:
        """Replace a field type from a Solr collection or core.

        Args:
            **field_kwargs: Any valid Solr field definition values, but
                must be a full redefinition, not a partial update.
        """
        self._post_field('replace-field-type', **field_kwargs)

    def get_schema(self) -> AttrDict:
        """Get the full schema for a Solr collection or core.

        Returns:
          Schema as returned by Solr.
        """
        response = self.make_request('get', self.url)
        if response:
            return response.schema

    def list_fields(self, fields: list=None ,
                    includeDynamic: bool=False,
                    showDefaults: bool=False) -> list:
        """Get a list of field definitions for a Solr Collection or core.

        Args:
          fields: A list of fields to filter by.
          includeDynamic: Include Solr dynamic fields in search.
          showDefaults: Show default Solr fields.

        Returns:
          list of fields as returned by Solr.
        """
        url = urljoin('%s/' % self.url, 'fields')
        params = {}
        if fields:
            params['fl'] = ','.join(fields)
        params['includeDynamic'] = includeDynamic
        params['showDefaults'] = showDefaults
        response = self.make_request('get', url, params=params)
        if response:
            return response.fields

    def list_copy_fields(self, source_fl: Optional[list]=None,
                         dest_fl: Optional[list]=None) -> List[AttrDict]:
        """Return a list of copy fields from Solr.

        Args:
            source_fl: Source field to filter by.
            dest_fl: Destination field to filter by.

        Returns:
            list of copy fields as returned by Solr.
        """
        url = urljoin('%s/' % self.url, 'copyfields')
        params = {}
        if source_fl:
            params['source.fl'] = ','.join(source_fl)
        if dest_fl:
            params['dest.fl'] = ','.join(dest_fl)
        response = self.make_request('get', url, params=params)
        if response:
            return response.copyFields

    def list_field_types(self, showDefaults: bool=True) -> List[AttrDict]:
        """List all field types in a Solr collection or core.

        Args:
            showDefaults: Show default fields

        Returns:
          list of copy fields as returned by Solr.
        """
        url = urljoin('%s/' % self.url, 'fieldtypes')
        params = {}
        params['showDefaults'] = showDefaults
        response = self.make_request('get', url, params=params)
        if response:
            return response.fieldTypes



from collections import OrderedDict
import logging
from typing import Any, Dict, Optional

from attrdict import AttrDict
import requests

from parasolr import __version__ as parasol_version
from parasolr.solr.base import ClientBase
from parasolr.solr.schema import Schema
from parasolr.solr.update import Update
from parasolr.solr.admin import CoreAdmin


logger = logging.getLogger(__name__)


## NOTE: As a rule, Solr parameters that are camelcased are retained that way
# despite not being hugely Pythonic, for consistency with Solr's responses
# and API documentation.

class ParasolrDict(AttrDict):
    """A subclass of :class:`attrdict.AttrDict` that can convert itself to a
    regular dictionary."""

    def as_dict(self):
        """Copy attributes from self as a dictionary, and recursively convert
        instances of :class:`ParasolrDict`."""
        copy = {}
        for k, v in self.items():
            if isinstance(v, ParasolrDict):
                copy[k] = v.as_dict()
            else:
                copy[k] = v
        return copy

class QueryResponse:
    """Thin wrapper to give access to Solr select responses.

    Args:
        response: A Solr query response
    """
    def __init__(self, response: Dict) -> None:
        # cast to ParasolrDict for any dict-like object
        response = ParasolrDict(response)
        self.numFound = int(response.response.numFound)
        self.start = int(response.response.start)
        self.docs = response.response.docs
        self.params = response.responseHeader.params
        self.facet_counts = {}
        if 'docs' in response.response:
            self.docs = response.response.docs
        if 'facet_counts' in response:
            self.facet_counts = \
                self._process_facet_counts(response.facet_counts)
        # NOTE: To access facet_counts.facet_fields or facet_counts.facet_ranges
        # as OrderedDicts, you must use dict notation (or AttrDict *will*
        # convert).

    def _process_facet_counts(self, facet_counts: AttrDict) \
            -> AttrDict:
        """Convert facet_fields and facet_ranges to OrderedDict.

        Args:
          facet_counts: Solr facet_counts field.

        Returns:
          Solr facet_counts field
        """
        if 'facet_fields' in facet_counts:
            for k, v in facet_counts.facet_fields.items():
                facet_counts['facet_fields'][k] = \
                    OrderedDict(zip(v[::2], v[1::2]))
        if 'facet_ranges' in facet_counts:
            for k, v in facet_counts.facet_ranges.items():
               facet_counts['facet_ranges'][k]['counts'] = \
                   OrderedDict(zip(v['counts'][::2], v['counts'][1::2]))
        return facet_counts


class SolrClient(ClientBase):
    """Class to aggregate all of the other Solr APIs and settings.

    Args:
        solr_url: Base url for solr.
        collection: Name of Solr collection or core.
        commitWithin: Time in ms for soft commits to happen.
        session: A python-requests :class:`requests.Session`.
    """

    #: CoreAdmin API handler
    core_admin_handler = 'admin/cores'
    #: Select handler
    select_handler = 'select'
    #: Schema API handler
    schema_handler = 'schema'
    #  Update API handler
    update_handler = 'update'
    #: core or collection name
    collection = ''
    #: commitWithin time in ms
    commitWithin = 1000


    def __init__(self, solr_url: str, collection: str,
                 commitWithin: Optional[int] = None,
                 session: Optional[requests.Session] = None) -> None:
        # Go ahead and create a session if one is not passed in
        super().__init__(session=session)

        self.solr_url = solr_url
        self.collection = collection
        if commitWithin:
            self.commitWithin = commitWithin
        self.session.headers = {
            'User-Agent': 'parasolr/%s (python-requests/%s)' % \
                (parasol_version, requests.__version__)
        }

        # attach remainder of API using a common session
        # and common settings
        self.schema = Schema(
            self.solr_url,
            self.collection,
            self.schema_handler,
            self.session
        )
        self.update = Update(
            self.solr_url,
            self.collection,
            self.update_handler,
            self.commitWithin
        )
        self.core_admin = CoreAdmin(
            self.solr_url,
            self.core_admin_handler,
            self.session)

    def query(self, wrap: bool = True, **kwargs: Any) -> Optional[QueryResponse]:
        """Perform a query with the specified kwargs.

        Args:
            **kwargs: Any valid Solr search parameters.

        Returns:
            A search QueryResponse.
        """
        url = self.build_url(self.solr_url, self.collection,
                             self.select_handler)
        # use POST for efficiency and send as x-www-form-urlencoded
        headers = {'content-type': 'application/x-www-form-urlencoded'}
        response = self.make_request(
            'post',
            url,
            headers=headers,
            params=kwargs
        )
        if response:
            # queries return the search response for now
            return QueryResponse(response) if wrap else response



from io import StringIO
from unittest.mock import Mock, patch

import pytest
import requests

try:
    from django.core.management import call_command
    from django.core.management.base import CommandError
    from django.test import override_settings

    from parasolr.management.commands import solr_schema, index
except ImportError:
    pass

from parasolr.tests.utils import skipif_no_django


@skipif_no_django
class TestSolrSchemaCommand:

    def test_report_changes(self):
        counts = {'added': 3, 'replaced': 0, 'deleted': 1}
        cmd = solr_schema.Command()
        with patch.object(cmd, 'stdout') as mock_stdout:
            cmd.report_changes(counts, 'field')

            assert mock_stdout.write.call_count == 2
            mock_stdout.write.assert_any_call('Added 3 fields')
            mock_stdout.write.assert_any_call('Deleted 1 field')

            # nothing to report
            counts = {'added': 0, 'replaced': 0, 'deleted': 0}
            mock_stdout.reset_mock()
            cmd.report_changes(counts, 'field')
            assert mock_stdout.write.call_count == 0

    @patch('parasolr.management.commands.solr_schema.SolrClient')
    def test_handle_connection_error(self, mocksolr):
        mocksolr.return_value.core_admin.ping.side_effect = \
            requests.exceptions.ConnectionError

        with pytest.raises(CommandError):
            solr_schema.Command().handle()

    @patch('parasolr.management.commands.solr_schema.SolrClient')
    @patch('parasolr.management.commands.solr_schema.SolrSchema')
    @patch('parasolr.management.commands.solr_schema.input')
    def test_handle_core(self, mockinput, mocksolrschema, mocksolrclient):
        # using mock SolrSchema to avoid exception on get_configuration

        mocksolr = mocksolrclient.return_value
        mocksolr.collection = 'test-coll'
        mocksolr.core_admin.ping.return_value = False

        cmd = solr_schema.Command()
        cmd.stdout = StringIO()

        # simulate user says no when asked to create core

        mockinput.return_value = 'n'
        cmd.handle()
        mockinput.assert_called_with(
            'Solr core %s does not exist. Create it? (y/n) ' %
            mocksolr.collection)
        mocksolr.core_admin.create.assert_not_called()

        # simulate user says yes when asked to create core
        mockinput.reset_mock()
        mockinput.return_value = 'Y'
        with override_settings(SOLR_CONNECTIONS=\
            {'default': {'CONFIGSET': 'test_config'}}):

            cmd.handle()
            # called once
            assert mockinput.call_count
            mocksolr.core_admin.create.assert_called_with(
                mocksolr.collection, configSet='test_config')

            # simulate no input requested
            mockinput.reset_mock()
            mocksolr.reset_mock()
            mocksolr.core_admin.ping.return_value = False
            cmd.handle(noinput=True)
            # input not called, but create should be called
            mockinput.assert_not_called()
            assert mocksolr.core_admin.create.call_count

            # should work the same way from the command line
            mockinput.reset_mock()
            mocksolr.reset_mock()
            mocksolr.core_admin.ping.return_value = False
            call_command('solr_schema', noinput=True, verbosity=0)
            # input not called, but create should be called
            mockinput.assert_not_called()
            assert mocksolr.core_admin.create.call_count

            # simulate collection does exist - no input or create
            mocksolr.reset_mock()
            mocksolr.core_admin.ping.return_value = True
            cmd.handle()
            mockinput.assert_not_called()
            mocksolr.core_admin.create.assert_not_called()

    @patch('parasolr.management.commands.solr_schema.SolrClient')
    @patch('parasolr.management.commands.solr_schema.SolrSchema')
    def test_handle_no_schema(self, mocksolrschema, mocksolrclient):
        mocksolr = mocksolrclient.return_value
        mocksolr.collection = 'test-coll'
        mocksolr.core_admin.ping.return_value = True
        err_msg = 'No Solr schema configuration found'
        mocksolrschema.get_configuration.side_effect = Exception(err_msg)

        with pytest.raises(CommandError) as err:
            solr_schema.Command().handle()
        assert err_msg in str(err)

    @patch('parasolr.management.commands.solr_schema.SolrClient')
    @patch('parasolr.management.commands.solr_schema.SolrSchema')
    def test_handle(self, mocksolrschema, mocksolrclient):
        mocksolr = mocksolrclient.return_value
        mocksolr.collection = 'test-coll'
        mocksolr.core_admin.ping.return_value = True

        schema_config = mocksolrschema.get_configuration.return_value

        cmd = solr_schema.Command()
        with patch.object(cmd, 'report_changes') as mock_report:
            cmd.handle()
            # should get schema config
            mocksolrschema.get_configuration.assert_called_with()
            # should configure field types and report
            schema_config.configure_fieldtypes.assert_called_with(mocksolr)
            mock_report.assert_any_call(
                schema_config.configure_fieldtypes.return_value, 'field type')
            # should configure fields and report
            schema_config.configure_fields.assert_called_with(mocksolr)
            mock_report.assert_any_call(
                schema_config.configure_fields.return_value, 'field')

            # should reload core
            mocksolr.core_admin.reload.assert_called_with(mocksolr.collection)


@skipif_no_django
class TestIndexCommand:

    @patch('parasolr.management.commands.index.Indexable')
    def test_index(self, mockindexable):
        # index data into solr and catch an error
        cmd = index.Command()
        cmd.solr = Mock()

        test_index_data = range(5)
        cmd.index(test_index_data)
        mockindexable.index_items.assert_called_with(test_index_data, progbar=None)

        # solr connection exception should raise a command error
        with pytest.raises(CommandError):
            mockindexable.index_items.side_effect = requests.exceptions.ConnectionError
            cmd.index(test_index_data)

    def test_clear(self):
        cmd = index.Command()
        cmd.solr = Mock()
        cmd.stdout = StringIO()

        cmd.clear('all')
        cmd.solr.update.delete_by_query.assert_called_with('*:*')

        cmd.solr.reset_mock()
        cmd.clear('work')
        cmd.solr.update.delete_by_query.assert_called_with('item_type:work')

        cmd.solr.reset_mock()
        cmd.clear('foo')
        cmd.solr.update.delete_doc_by_query.assert_not_called()

        cmd.stdout = StringIO()
        cmd.verbosity = 3
        cmd.clear('works')
        assert cmd.stdout.getvalue() == 'Clearing works from the index'

        cmd.stdout = StringIO()
        cmd.clear('all')
        assert cmd.stdout.getvalue() == 'Clearing everything from the index'

        # should also work from the command line
        with patch('parasolr.management.commands.index.SolrClient'):
            cmd.stdout.seek(0)
            call_command('index', index='none', clear='all', stdout=cmd.stdout)
            assert 'Clearing everything from the index' in cmd.stdout.getvalue()
            assert 'Indexed 0 items' in cmd.stdout.getvalue()

    @patch('parasolr.management.commands.index.Indexable')
    @patch('parasolr.management.commands.index.SolrClient')
    def test_handle_index_by_id(self, mocksolr, mockindexable):
        # create a mock indexable subclass to be returned by
        # mockindexable
        mock_indexable_model = Mock()
        mock_indexable_model.index_item_type.return_value = 'simple'
        mockindexable.all_indexables.return_value = [
            mock_indexable_model
        ]
        mockindexable.ID_SEPARATOR = '.'

        # patch the method that actually does the indexing (tested elsewhere)
        with patch.object(index.Command, 'index') as mock_index_meth:
            cmd = index.Command()
            cmd.stdout = StringIO()

            mock_index_meth.return_value = 1
            cmd.init_indexables()
            cmd.handle(index_ids=['simple.a', 'simple.b'], clear=None, no_progress=True)
            mock_indexable_model.objects.get.assert_any_call(pk='a')
            mock_indexable_model.objects.get.assert_any_call(pk='b')
            # items are indexed together in a single batch
            assert mock_index_meth.call_count == 1

            # handle id with unknown index label
            mock_index_meth.reset_mock()
            with pytest.raises(CommandError) as err:
                cmd.handle(index_ids=['foo.1'], clear=None, no_progress=True)
            assert not mock_index_meth.call_count
            assert "Unrecognized index id 'foo.1'" in str(err)

            # handle id without separator
            mock_index_meth.reset_mock()
            with pytest.raises(CommandError) as err:
                cmd.handle(index_ids=['foo:1'], clear=None, no_progress=True)
            assert not mock_index_meth.call_count
            assert "Unrecognized index id 'foo:1'" in str(err)

    @patch('parasolr.management.commands.index.progressbar')
    @patch('parasolr.management.commands.index.SolrClient')
    @patch('parasolr.indexing.SolrClient')
    def test_call_command(self, mocksolr, mocksolr2, mockprogbar):
        mocksolr = Mock()

        # patch the method that actually does the indexing (tested elsewhere)
        with patch.object(index.Command, 'index') as mock_index_meth:
            stdout = StringIO()
            # index method returns number of items indexed
            mock_index_meth.return_value = 6

            # index all indexable content
            call_command('index', index='all', stdout=stdout)
            # should be called once for each indexable subclass
            assert mock_index_meth.call_count == 2
            # call order is not guaranteed, not inspecting here
            # commit called after works are indexed
            mocksolr2.return_value.update.index.assert_called_with([], commit=True)
        # self.solr.update.index([], commit=True)
            # mocksolr.commit.assert_called_with(test_coll, openSearcher=True)

            # progressbar should be initialized and finished
            mockprogbar.ProgressBar.assert_called_with(
                redirect_stdout=True, max_value=6)
            mockprogbar.ProgressBar.return_value.finish.assert_called_with()

            # request no progress bar
            mockprogbar.reset_mock()
            call_command('index', index='all', no_progress=True, stdout=stdout)
            mockprogbar.ProgressBar.assert_not_called()

            # request single type; not enough data to run progress bar
            mockprogbar.reset_mock()
            mock_index_meth.reset_mock()
            call_command('index', index='simple', no_progress=True, stdout=stdout)
            mockprogbar.ProgressBar.assert_not_called()
            # should only be called once
            assert mock_index_meth.call_count == 1

            # index nothing
            mock_index_meth.reset_mock()
            call_command('index', index='none', stdout=stdout)
            assert not mock_index_meth.call_count



from unittest.mock import patch, Mock

from attrdict import AttrDict
import pytest

from parasolr.solr import SolrClient
from parasolr import schema


def test_solr_fields():
    """unit test for solr field descriptor"""

    # create a class with fields to test field descriptors
    class TestySolrFields:
        mystring = schema.SolrStringField()
        required_str = schema.SolrStringField(required=True)
        multival_str = schema.SolrStringField(multivalued=True)
        custom_field = schema.SolrField('text_en')

    # get on the field returns solr config info
    assert TestySolrFields.mystring ==  \
        {'type': 'string', 'required': False, 'multiValued': False}
    assert TestySolrFields.required_str ==  \
        {'type': 'string', 'required': True, 'multiValued': False}
    assert TestySolrFields.multival_str ==  \
        {'type': 'string', 'required': False, 'multiValued': True}
    assert TestySolrFields.custom_field ==  \
        {'type': 'text_en', 'required': False, 'multiValued': False}

    # explicitly read-only
    with pytest.raises(AttributeError):
        TestySolrFields().mystring = 'foo'

class TestAnalyzer(schema.SolrAnalyzer):
    """ test analyzer for checking analyzer and field type logic"""
    tokenizer = 'solr.StandardTokenizerFactory'
    filters = [
        {"class": "solr.StopFilterFactory", "ignoreCase": True,
         "words": "lang/stopwords_en.txt"},
        {"class": "solr.LowerCaseFilterFactory"},
    ]


def test_analyzer():
    analyzer_conf = TestAnalyzer.as_solr_config()
    assert analyzer_conf['tokenizer']['class'] == TestAnalyzer.tokenizer
    assert analyzer_conf['filters'] == TestAnalyzer.filters


def test_solr_field_types():
    """unit test for solr field type descriptor and analyzer"""

    # create a class with fields to test field descriptors
    class TestSchema:
        # field type declaration
        text_en = schema.SolrFieldType('solr.TextField',
                                       analyzer=TestAnalyzer)

    # get on the field type returns solr config details
    field_config = TestSchema.text_en
    assert field_config['class'] == 'solr.TextField'
    assert field_config['analyzer'] == TestAnalyzer.as_solr_config()

    # explicitly read-only
    with pytest.raises(AttributeError):
        TestSchema().text_en = 'foo'


class TestSchema:

    def test_get_configuration__none(self):
        # error if no subclass defined
        with pytest.raises(Exception) as err:
            schema.SolrSchema.get_configuration()
        assert 'No Solr schema configuration found' in str(err)

    def test_get_configuration__one(self):
        # define one subclass
        class SchemaOne(schema.SolrSchema):
            pass
        assert schema.SolrSchema.get_configuration() == SchemaOne

    def test_get_configuration__multiple(self):
        # define more than one subclass
        class SchemaOne(schema.SolrSchema):
            pass

        class SchemaTwo(schema.SolrSchema):
            pass

        with pytest.raises(Exception) as err:
            schema.SolrSchema.get_configuration()
        assert 'Currently only one Solr schema configuration is supported' in str(err)

    def test_get_field_names(self):

        class LocalTestSchema(schema.SolrSchema):
            name = schema.SolrStringField(required=True)
            title = schema.SolrStringField()
            date = schema.SolrField('date')

        schemafields = LocalTestSchema.get_field_names()
        for field in ['name', 'title', 'date']:
            assert field in schemafields

    def test_get_field_types(self):

        class TestAnalyzer(schema.SolrAnalyzer):
            pass

        class LocalTestSchema(schema.SolrSchema):
            text_en = schema.SolrFieldType('solr.TextField', analyzer=TestAnalyzer)
            text_sp = schema.SolrFieldType('solr.TextField', analyzer=TestAnalyzer)

        field_types = LocalTestSchema.get_field_types()
        for field_type in ['text_en', 'text_sp']:
            assert field_type in field_types

    def test_configure_fields(self):
        mocksolr = Mock()

        class LocalTestSchema(schema.SolrSchema):
            name = schema.SolrStringField(required=True)
            title = schema.SolrStringField()
            date = schema.SolrField('date')


        with patch.object(LocalTestSchema, 'configure_copy_fields') \
          as mock_config_cp_fields:

            # simulate only standard fields defined
            mocksolr.schema.list_fields.return_value = [
                AttrDict({'name': 'id'}), AttrDict({'name': '_version_'})
            ]

            result = LocalTestSchema.configure_fields(mocksolr)
            assert result.added == 3
            assert result.replaced == 0
            assert result.deleted == 0
            assert mock_config_cp_fields.call_count == 1

            # add field with field config options should be called for each
            for field in ['name', 'title', 'date']:
                mocksolr.schema.add_field.assert_any_call(
                    name=field, **getattr(LocalTestSchema, field))

            mocksolr.schema.replace_field.assert_not_called()
            mocksolr.schema.delete_field.assert_not_called()

            # simulate all standard fields defined and one extra
            mocksolr.reset_mock()
            mocksolr.schema.list_fields.return_value = [
                AttrDict({'name': 'name'}),
                AttrDict({'name': 'title'}),
                AttrDict({'name': 'date'}),
                AttrDict({'name': 'foobar'})
            ]

            result = LocalTestSchema.configure_fields(mocksolr)
            assert result.added == 0
            assert result.replaced == 3
            assert result.deleted == 1
            mocksolr.schema.add_field.assert_not_called()
            # replace field with field config options should be called for each
            for field in ['name', 'title', 'date']:
                mocksolr.schema.replace_field.assert_any_call(
                    name=field, **getattr(LocalTestSchema, field))

            mocksolr.schema.delete_field.assert_called_with('foobar')

    def test_configure_copy_fields(self):
        mocksolr = Mock()

        class LocalTestSchema(schema.SolrSchema):
            #: copy fields, e.g. for facets or default search field ('text')
            copy_fields = {
                'author': ['text', 'author_exact'],
                'collections': 'collections_s',
                'title': ['title_nostem', 'title_s'],
                'subtitle': 'subtitle_s',
            }

        # simulate no copy fields
        mocksolr.schema.list_copy_fields.return_value = []
        LocalTestSchema.configure_copy_fields(mocksolr)
        # should be called once for each
        for source, dest in LocalTestSchema.copy_fields.items():
            mocksolr.schema.add_copy_field.assert_any_call(source, dest)
        # delete not called
        mocksolr.schema.delete_copy_field.assert_not_called()

        # simulate some to not be added, some to remove
        mocksolr.reset_mock()
        mocksolr.schema.list_copy_fields.return_value = [
            # valid source and destination in list
            AttrDict({'source': 'author', 'dest': 'text'}),
            # valid source, destination not in list
            AttrDict({'source': 'author', 'dest': 'foobar'}),
            # valid source & destination
            AttrDict({'source': 'collections', 'dest': 'collections_s'}),
            # invalid source
            AttrDict({'source': 'foo', 'dest': 'bar'}),
            # valid source, destination not == value
            AttrDict({'source': 'subtitle', 'dest': 'baz'}),
        ]

        LocalTestSchema.configure_copy_fields(mocksolr)
        # should be called for title, subtitle, and second author dest
        assert mocksolr.schema.add_copy_field.call_count == 3

        # extra author copy dest should be removed
        mocksolr.schema.delete_copy_field.assert_any_call('author', 'foobar')
        # extra source/dest should be removed
        mocksolr.schema.delete_copy_field.assert_any_call('foo', 'bar')
        mocksolr.schema.delete_copy_field.assert_any_call('subtitle', 'baz')
        assert mocksolr.schema.delete_copy_field.call_count == 3

    def test_configure_field_types(self):
        mocksolr = Mock()

        class TestAnalyzer(schema.SolrAnalyzer):
            pass

        class LocalTestSchema(schema.SolrSchema):
            text_en = schema.SolrFieldType('solr.TextField', analyzer=TestAnalyzer)

        # simulate no field types defined
        mocksolr.schema.list_field_types.return_value = []
        # current_field_types = {ftype['name']: ftype
                   # for ftype in solr.schema.list_field_types()}
            # simulate only standard fields defined

        result = LocalTestSchema.configure_fieldtypes(mocksolr)
        assert result.added == 1
        assert result.updated == 0

        mocksolr.schema.add_field_type.assert_called_with(
            name='text_en', **LocalTestSchema.text_en)
        mocksolr.schema.replace_field_type.assert_not_called()

        # simulate field type already defined
        mocksolr.reset_mock()
        mocksolr.schema.list_field_types.return_value = [
            {'name': 'text_en'}
        ]
        result = LocalTestSchema.configure_fieldtypes(mocksolr)
        assert result.added == 0
        assert result.updated == 1
        mocksolr.schema.replace_field_type.assert_called_with(
            name='text_en', **LocalTestSchema.text_en)
        mocksolr.schema.add_field_type.assert_not_called()

        # no field types defined - should not error, do nothing
        assert not schema.SolrSchema.configure_fieldtypes(mocksolr)



from collections import OrderedDict
from unittest.mock import patch, Mock

import pytest

from parasolr.solr import SolrClient
from parasolr.solr.client import QueryResponse, ParasolrDict
from parasolr.query import SolrQuerySet


class TestSolrQuerySet:

    def test_init(self):
        # use solr client if passed in
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)
        assert sqs.solr == mocksolr

    def test_query_opts(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)

        # default behavior: find all starting at 0
        query_opts = sqs.query_opts()
        assert query_opts['start'] == 0
        assert query_opts['q'] == '*:*'
        # don't include unset options
        for opt in ['fq', 'rows', 'sort', 'fl', 'hl', 'hl.field', 'facet']:
            assert opt not in query_opts

        # customized query opts
        sqs.start = 10
        sqs.stop = 20
        sqs.sort_options = ['title asc', 'date desc']
        sqs.filter_qs = ['item_type:work']
        sqs.search_qs = ['title:reading', 'author:johnson']
        sqs.field_list = ['title', 'author', 'date:pubyear_i']
        sqs.highlight_field = 'content'
        sqs.highlight_opts = {'snippets': 3, 'method': 'unified'}
        sqs.facet_field = ['item_type', 'member_type']
        sqs.facet_opts = {'sort': 'count'}
        query_opts = sqs.query_opts()

        assert query_opts['start'] == sqs.start
        assert query_opts['rows'] == sqs.stop - sqs.start
        assert query_opts['fq'] == sqs.filter_qs
        assert query_opts['q'] == ' AND '.join(sqs.search_qs)
        assert query_opts['sort'] == ','.join(sqs.sort_options)
        assert query_opts['fl'] == ','.join(sqs.field_list)
        # highlighting should be turned on
        assert query_opts['hl']
        assert query_opts['hl.field'] == 'content'
        # highlighting options added with hl.prefix
        assert query_opts['hl.snippets'] == 3
        assert query_opts['hl.method'] == 'unified'
        # make sure faceting opts are preserved
        assert query_opts['facet'] is True
        assert query_opts['facet.field'] == sqs.facet_field
        assert query_opts['facet.sort'] == 'count'

    def test_query(self):
        mocksolr = Mock(spec=SolrClient)
        mocksolr.query.return_value.docs = []
        sqs = SolrQuerySet(mocksolr)
        query_sqs = sqs.query()
        # returns a copy, not same queryset
        assert query_sqs != sqs
        # sets the result cache (via get_results)
        assert query_sqs._result_cache
        # result cache not set on original
        assert not sqs._result_cache


    def test_get_results(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)

        mockresponse = Mock()
        mockresponse.docs = [
            ParasolrDict({'a': 1})
        ]
        mocksolr.query.return_value = mockresponse

        # by default, should query solr with options from query_opts
        # and wrap = false
        query_opts = sqs.query_opts()
        assert sqs.get_results() == mockresponse.docs
        assert sqs._result_cache == mockresponse
        mocksolr.query.assert_called_with(**query_opts)

        # parameters passed in take precedence
        local_opts = {'q': 'name:hemingway', 'sort': 'name asc'}
        sqs.get_results(**local_opts)
        # update copy of query opts with locally passed in params
        query_opts.update(local_opts)
        mocksolr.query.assert_called_with(**query_opts)

    def test_count(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)

        # simulate result cache already populated; should use
        sqs._result_cache = Mock()
        sqs._result_cache.numFound = 5477
        assert sqs.count() == 5477
        mocksolr.query.assert_not_called()

        # if cache is not populated, should query for count
        sqs._result_cache = None
        mocksolr.query.return_value = Mock()
        mocksolr.query.return_value.numFound = 343
        assert sqs.count() == 343
        count_query_opts = sqs.query_opts()
        count_query_opts['rows'] = 0
        count_query_opts['hl'] = False
        count_query_opts['facet'] = False
        mocksolr.query.assert_called_with(**count_query_opts)
        # cache should not be populated
        assert not sqs._result_cache

    @patch('parasolr.query.QueryResponse')
    def test_get_facets(self, mockQR):
        mocksolr = Mock(spec=SolrClient)
        # mock cached solr response
        mock_response = Mock()
        sqs = SolrQuerySet(mocksolr)
        # mock out return of MockQR constructor to ensure it calls
        # facet_counts.facet_fields
        mockQR.return_value = Mock()
        mockQR.return_value.facet_counts = {'facet_fields': OrderedDict(a=1)}
        sqs._result_cache = mock_response

        ret = sqs.get_facets()
        # QueryResponse called to wrap mock_response
        assert mockQR.called
        # called with the cached response
        mockQR.assert_called_with(mock_response)
        # casts return to an OrderedDict
        assert isinstance(ret, OrderedDict)
        # return the value of facet_counts.facet_fields
        assert ret == OrderedDict(a=1)

        # now test no cached result
        mocksolr.query.return_value = Mock()
        mocksolr.query.return_value.facet_counts = {'facet_fields': OrderedDict(b=2)}
        sqs._result_cache = None
        # clear the previous mocks
        mockQR.reset_mock()

        ret = sqs.get_facets()
        # QueryResponse not called to wrap return of query
        assert not mockQR.called
        # solr.query called
        assert mocksolr.query.called
        # should be called with rows=0 and hl=False to avoid inefficiencies
        name, args, kwargs = mocksolr.query.mock_calls[0]
        assert kwargs['rows'] == 0
        assert not kwargs['hl']
        # casts return to an OrderedDict
        assert isinstance(ret, OrderedDict)
        # return should be the return value of facet_counts.facet_fields
        assert ret == OrderedDict(b=2)

    def test_filter(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)

        # arg options added to filter list as is
        new_filters = ['item_type:work', 'date:[1550 TO 1900]']
        filtered_qs = sqs.filter(*new_filters)
        # returned queryset has the filters
        assert filtered_qs.filter_qs == new_filters
        # original queryset is unchanged
        assert not sqs.filter_qs

        # keyword arg options converted into filters
        filtered_qs = sqs.filter(item_type='work', date=1500)
        # returned queryset has the filters
        assert 'item_type:work' in filtered_qs.filter_qs
        assert 'date:1500' in filtered_qs.filter_qs
        # original queryset is unchanged
        assert not sqs.filter_qs

        # chaining adds to the filters
        filtered_qs = sqs.filter(item_type='work').filter(date=1500) \
                         .filter('name:he*')
        assert 'item_type:work' in filtered_qs.filter_qs
        assert 'date:1500' in filtered_qs.filter_qs
        assert 'name:he*' in filtered_qs.filter_qs

    def test_facet(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)
        # facet a search
        facet_list = ['person_type', 'item_type']
        faceted_qs = sqs.facet(*facet_list)
        # faceting should be set on
        assert faceted_qs.facet_field == facet_list
        # facet opts and field for original queryset should be unchanged
        assert not sqs.facet_opts
        assert not sqs.facet_field

        # a call to another method should leave facet options as is
        faceted_qs = faceted_qs.filter(foo='bar')
        assert faceted_qs.facet_field== facet_list
        # subsequents calls to facet should simply reset list
        facet_list = ['foobars']
        faceted_qs = faceted_qs.facet(*facet_list)
        assert faceted_qs.facet_field == facet_list
        # kwargs should simply be set in facet opts
        faceted_qs = faceted_qs.facet(*facet_list, sort='count')
        assert faceted_qs.facet_field == facet_list
        assert faceted_qs.facet_opts['sort'] == 'count'

    def test_search(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)

        # arg options added to filter list as is
        queries = ['item_type:work', 'date:[1550 TO *]']
        search_sqs = sqs.search(*queries)
        # returned queryset has the filters
        assert search_sqs.search_qs == queries
        # original queryset is unchanged
        assert not sqs.search_qs

        # keyword arg options are converted
        search_sqs = sqs.search(item_type='work', date=1550)
        assert 'item_type:work' in search_sqs.search_qs
        assert 'date:1550' in search_sqs.search_qs
        # original queryset is unchanged
        assert not sqs.search_qs

        # chaining is additive
        search_sqs = sqs.search(item_type='work').search(date=1550)
        assert 'item_type:work' in search_sqs.search_qs
        assert 'date:1550' in search_sqs.search_qs
        # original queryset is unchanged
        assert not sqs.search_qs

    def test_order_by(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)

        # default is ascending
        sorted_sqs = sqs.order_by('date')
        assert sorted_sqs.sort_options == ['date asc']
        # original queryset is unchanged
        assert not sqs.sort_options

        # - indicates descending
        sorted_sqs = sqs.order_by('-date')
        assert sorted_sqs.sort_options == ['date desc']
        # original queryset is unchanged
        assert not sqs.sort_options

        # can handle multiple
        sorted_sqs = sqs.order_by('title', 'date')
        assert sorted_sqs.sort_options == ['title asc', 'date asc']
        # original queryset is unchanged
        assert not sqs.sort_options

        # chaining works the same way
        sorted_sqs = sqs.order_by('title').order_by('date')
        assert sorted_sqs.sort_options == ['title asc', 'date asc']
        # original queryset is unchanged
        assert not sqs.sort_options

    def test_only(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)
        only_fields = ['title', 'author', 'date']
        # field name only, single list
        fields_sqs = sqs.only(*only_fields)
        # field list refined
        assert fields_sqs.field_list == only_fields
        # original field list unchanged
        assert not sqs.field_list

        # chaining is not equivalent, but *replaces*
        fields_sqs = sqs.only('title').only('author')
        # field list refined
        assert fields_sqs.field_list == ['author']
        # original field list unchanged
        assert not sqs.field_list

        # only with field alias
        fields_sqs = sqs.only(title='title_i')
        assert 'title:title_i' in fields_sqs.field_list

    def test_highlight(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)
        # field only, defaults
        highlight_qs = sqs.highlight('content')
        assert highlight_qs.highlight_field == 'content'
        assert highlight_qs.highlight_opts == {}
        # original unchanged
        assert sqs.highlight_field is None

        # field and opts
        highlight_qs = sqs.highlight('text', snippets=3, method='unified')
        assert highlight_qs.highlight_field == 'text'
        assert highlight_qs.highlight_opts == \
            {'snippets': 3, 'method': 'unified'}
        # original unchanged
        assert sqs.highlight_field is None
        assert sqs.highlight_opts == {}

    def test_raw_query_parameters(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)
        raw_q = {'extra_query': 'foobar'}
        raw_sqs = sqs.raw_query_parameters(**raw_q)
        # raw query stored
        assert raw_sqs.raw_params == raw_q
        # included in query opts
        assert raw_q['extra_query'] in raw_sqs.query_opts()['extra_query']
        # original unchanged
        assert sqs.raw_params == {}

        # additional raw params add rather than replace
        rawer_sqs = raw_sqs.raw_query_parameters(another='two')
        assert len(rawer_sqs.raw_params) == 2

    def test_get_highlighting(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)
        # simulate result cache already populated, no highlighting
        sqs._result_cache = {'response': {'docs': []}}
        assert sqs.get_highlighting() == {}

        # simulate response with highlighting
        mock_highlights = {'id1': {'text': ['sample match content']}}
        sqs._result_cache = {'highlighting': mock_highlights}
        assert sqs.get_highlighting() == mock_highlights

        # should populate cache if empty
        sqs._result_cache = None
        with patch.object(sqs, 'get_results') as mock_get_results:
            def set_result_cache():
                sqs._result_cache = Mock()
            mock_get_results.side_effect = set_result_cache

            sqs.get_highlighting()
            mock_get_results.assert_called_with()

    def test_all(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)
        # all just calls clone and returns the queryset copy
        with patch.object(sqs, '_clone') as mockclone:
            all_sqs = sqs.all()
            mockclone.assert_called_with()
            assert all_sqs == mockclone.return_value

    def test_none(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)
        none_sqs = sqs.none()
        # new queryset search replaced with something to return nothing
        assert none_sqs.search_qs == ['NOT *:*']
        # original queryset unchanged
        assert sqs.search_qs == []

        # none after search terms replaces the search
        search_sqs = SolrQuerySet(mocksolr).search('item_type:work')
        none_sqs = search_sqs.none()
        assert search_sqs.search_qs == ['item_type:work']
        assert none_sqs.search_qs == ['NOT *:*']

    def test__clone(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)
        # clone default with no filters
        cloned_sqs = sqs._clone()
        assert cloned_sqs.start == 0
        assert cloned_sqs.stop is None
        assert cloned_sqs.search_qs == []
        assert cloned_sqs.filter_qs == []
        assert cloned_sqs.sort_options == []

        # set everything
        custom_sqs = sqs.filter(item_type='person').search(name='he*') \
                        .order_by('birth_year')
        custom_sqs.set_limits(10, 100)
        custom_clone = custom_sqs._clone()
        assert custom_clone.start == 10
        assert custom_clone.stop == 100
        # list fields should be equal but not be the same object
        assert custom_clone.search_qs == custom_sqs.search_qs
        assert not custom_clone.search_qs is custom_sqs.search_qs
        assert custom_clone.filter_qs == custom_sqs.filter_qs
        assert not custom_clone.filter_qs is custom_sqs.filter_qs
        assert custom_clone.sort_options == custom_sqs.sort_options
        assert not custom_clone.sort_options is custom_sqs.sort_options

        # subclass clone should return subclass

        class CustomSolrQuerySet(SolrQuerySet):
            pass

        custom_clone = CustomSolrQuerySet(mocksolr)._clone()
        assert isinstance(custom_clone, CustomSolrQuerySet)

        # sanity-check chaining
        sqs = SolrQuerySet(mocksolr)
        filtered_sqs = sqs.filter(item_type='person')
        # filter should be set
        assert 'item_type:person' in filtered_sqs.filter_qs
        sorted_sqs = filtered_sqs.order_by('birth_year')
        # sort and filter should be set
        assert 'item_type:person' in sorted_sqs.filter_qs
        assert 'birth_year asc' in sorted_sqs.sort_options
        search_sqs = sorted_sqs.search(name='hem*')
        # search , sort, and filter should be set
        assert 'item_type:person' in search_sqs.filter_qs
        assert 'birth_year asc' in search_sqs.sort_options
        assert 'name:hem*' in search_sqs.search_qs

    def test__lookup_to_filter(self):
        assert SolrQuerySet._lookup_to_filter('item_type', 'work') == \
            'item_type:work'

    def test_iter(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)
        with patch.object(sqs, 'get_results') as mock_get_results:
            mock_get_results.return_value = [{'id': 1}, {'id': 2}]
            results_iterator = sqs.__iter__()
            # not sure how best to test iterator...
            assert list(results_iterator) == mock_get_results.return_value
            mock_get_results.assert_called_with()

        # TODO: test iterating over a slice?

    def test_bool(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)
        with patch.object(sqs, 'get_results') as mock_get_results:
            # with results
            mock_get_results.return_value = [{'id': 1}, {'id': 2}]
            assert sqs

            # with no results
            mock_get_results.return_value = []
            assert not sqs

    def test_set_limits(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)
        sqs.set_limits(10, 50)
        assert sqs.start == 10
        assert sqs.stop == 50

        sqs.set_limits(None, None)
        assert sqs.start == 0
        assert sqs.stop is None

    def test_get_item(self):
        mocksolr = Mock(spec=SolrClient)
        sqs = SolrQuerySet(mocksolr)

        # simulate result cache already populated
        sqs._result_cache = Mock()
        sqs._result_cache.docs = [1, 2, 3, 4, 5]
        # single item
        assert sqs[0] == 1
        assert sqs[1] == 2
        # slice
        assert sqs[0:2] == [1, 2]
        # slice with step
        assert sqs[1:5:2] == [2, 4]

        # simulate result cache *not* populated
        sqs._result_cache = None
        # - slice
        sliced_qs = sqs[10:20]
        assert isinstance(sliced_qs, SolrQuerySet)
        assert sliced_qs.start == 10
        assert sliced_qs.stop == 20
        # - slice with implicit start
        sliced_qs = sqs[:5]
        assert isinstance(sliced_qs, SolrQuerySet)
        assert sliced_qs.start == 0
        assert sliced_qs.stop == 5
        # - slice with implicit end
        sliced_qs = sqs[3:]
        assert isinstance(sliced_qs, SolrQuerySet)
        assert sliced_qs.start == 3
        assert sliced_qs.stop is None

        # - slice with step - requires fetching results
        with patch.object(sqs, '_clone') as mock_clone:
            # for step slicing, calls list on the cloned queryset,
            # which calls iterate; supply list of numbers as response
            mock_clone.return_value.__iter__.return_value = range(20)
            sliced_qs = sqs[0:10:2]
            assert not isinstance(sliced_qs, SolrQuerySet)
            assert sliced_qs[1] == 2
            assert sliced_qs[-1] == 18

        # - single item
        with patch.object(sqs, '_clone') as mock_clone:
            mock_clone.return_value.get_results.return_value \
                = ['d']
            item = sqs[3]

            mock_clone.assert_called_with()
            mock_clone.return_value.set_limits.assert_called_with(3, 4)
            mock_clone.return_value.get_results.assert_called_with()
            assert item == 'd'

        # handle invalid input
        with pytest.raises(TypeError):
            assert sqs['foo']

        with pytest.raises(AssertionError):
            assert sqs[-1]

        with pytest.raises(AssertionError):
            assert sqs[:-1]



"""
Test pytest plugin fixture for django
"""
import uuid

import pytest
try:
    from django.conf import settings
    from django.test import override_settings

    from parasolr.pytest_plugin import get_test_solr_config

except ImportError:
    pass

from parasolr.tests.utils import skipif_no_django


pytest_plugins = "pytester"


@skipif_no_django
class TestGetTestSolrConfig:

    def test_no_testconfig(self):
        no_test_cfg = {
            'default': {
                'URL': 'http://example.com:8984/solr',
                'COLLECTION': 'mycoll'
            }
        }
        with override_settings(SOLR_CONNECTIONS=no_test_cfg):
            test_cfg = get_test_solr_config()
            assert test_cfg['URL'] == no_test_cfg['default']['URL']
            assert test_cfg['COLLECTION'] == \
                'test_%s' % no_test_cfg['default']['COLLECTION']

    def test_has_testconfig(self):
        cfg_with_test_cfg = {
            'default': {
                'URL': 'http://example.com:8984/solr',
                'COLLECTION': 'mycoll',
                'TEST': {
                    'URL': 'http://testing.example.com:8985/solr',
                    'COLLECTION': 'testing',
                    'COMMITWITHIN': 350
                }
            }
        }
        with override_settings(SOLR_CONNECTIONS=cfg_with_test_cfg):
            test_cfg = get_test_solr_config()
            assert test_cfg['URL'] == \
                cfg_with_test_cfg['default']['TEST']['URL']
            assert test_cfg['COLLECTION'] == \
                cfg_with_test_cfg['default']['TEST']['COLLECTION']
            assert test_cfg['COMMITWITHIN'] == \
                cfg_with_test_cfg['default']['TEST']['COMMITWITHIN']


@skipif_no_django
def test_configure_django_test_solr(testdir):
    """Sanity check pytest solr fixture."""

    solr_url = settings.SOLR_CONNECTIONS['default']['URL']
    solr_test_collection = settings.SOLR_CONNECTIONS['default']['TEST']['COLLECTION']
    solr_commit_within = settings.SOLR_CONNECTIONS['default']['TEST']['COMMITWITHIN']

    # NOTE: can't figure out how to get the plugin test to use
    # test-local test settings, so testing against project
    # testsettings for now

    # create a temporary pytest test file
    testdir.makepyfile(
        """
        from parasolr.django import SolrClient

        # causes "Plugin already registered" error on travis...
        # pytest_plugins = "parasolr.pytest_plugin"

        def test_solr_client():
            solr = SolrClient()
            # solr client should use test config
            assert solr.solr_url == '%s'
            test_collection = '%s'
            assert solr.collection == test_collection
            # should get test config for commit within
            assert solr.commitWithin == %d
            # core should exist
            assert solr.core_admin.status(core=test_collection)

    """ % (solr_url, solr_test_collection, solr_commit_within)
    )

    # run all tests with pytest with all pytest-django plugins turned off
    # result = testdir.runpytest('-p', 'no:django')
    result = testdir.runpytest_subprocess('--capture', 'no') #, '-p', 'no:django')
    # check that test case passed
    result.assert_outcomes(passed=1)



"""
Model-based indexing with Solr.

Items to be indexed in Solr should extend :class:`Indexable`. The
default implementation should work for most Django models; at a minimum
you should extend :meth:`Indexable.index_data` to include the information
to be indexed in Solr. You may also customize :meth:`Indexable.index_item_type`
and :meth:`Indexable.index_item_id`.

To manually index content in Solr, see
:mod:`~parasolr.management.commands.index` manage command documentation.

-------------------------

"""

import itertools
import logging

try:
    from django.db.models.query import QuerySet

    from parasolr.django import SolrClient
except ImportError:
    QuerySet = SolrClient = None


logger = logging.getLogger(__name__)


class Indexable:
    """Mixin for objects that are indexed in Solr.  Subclasses must implement
    `index_id` and `index` methods.
    """

    # NOTE: current implementation is Django-specific, intended for
    # use with django models. Should be possible to generalize once
    # we have other use cases.

    #: number of items to index at once when indexing a large number of items
    index_chunk_size = 150

    #: solr connection
    solr = None

    #: id separator for auto-generated index ids
    ID_SEPARATOR = '.'

    def __init__(self):
        # initialize connection to solr on first instance initialization
        Indexable._init_solr()

    @classmethod
    def _init_solr(cls):
        # store on the class to take advantage of sessions
        if cls.solr is None:
            cls.solr = SolrClient()

    @classmethod
    def all_indexables(cls):
        """Find all :class:`Indexable` subclasses for indexing."""
        return cls.__subclasses__()

    @classmethod
    def index_item_type(cls):
        """Label for this kind of indexable item. Must be unique
        across all Indexable items in an application. By default, uses
        Django model verbose name. Used in default index id and
        in index manage command. """
        return cls._meta.verbose_name

    def index_id(self):
        """Solr identifier. By default, combines :meth:`index item_type`
        and :attr:`id` with :attr:ID_SEPARATOR`."""
        return '%s%s%s' % (self.index_item_type(), self.ID_SEPARATOR,
                           self.id)

    def index_data(self):
        """Dictionary of data to index in Solr for this item.
        Default implementation adds  :meth:`index_id` and
        :meth:`index_item_type` """
        return {
            'id': self.index_id(),
            'item_type': self.index_item_type()
        }

    def index(self):
        """Index the current object in Solr.  Allows passing in
        parameter, e.g. to set a `commitWithin` value.
        """
        self.solr.update.index([self.index_data()])

    @classmethod
    def index_items(cls, items, progbar=None):
        """Indexable class method to index multiple items at once.  Takes a
        list, queryset, or generator of Indexable items or dictionaries.
        Items are indexed in chunks, based on :attr:`Indexable.index_chunk_size`.

        Args:
            items: list, queryset, or generator of indexable objects or dictionaries
            progbar: optional :class:`progressbar.Progressbar` object to
            update when indexing items in chunks.

        Returns:
            Total number of items indexed
        """

        # make sure solr client is initialized
        Indexable._init_solr()

        # if this is a queryset, use iterator to get it in chunks
        if QuerySet and isinstance(items, QuerySet):
            items = items.iterator()

        # if this is a normal list, convert it to an iterator
        # so we don't iterate the same slice over and over
        elif isinstance(items, list):
            items = iter(items)

        # index in chunks to support efficiently indexing large numbers
        # of items (adapted from index script)
        chunk = list(itertools.islice(items, cls.index_chunk_size))
        count = 0
        while chunk:
            # call index data method if present; otherwise assume item is dict
            cls.solr.update.index(
                [i.index_data() if hasattr(i, 'index_data') else i
                 for i in chunk])
            count += len(chunk)
            # update progress bar if one was passed in
            if progbar:
                progbar.update(count)

            # get the next chunk
            chunk = list(itertools.islice(items, cls.index_chunk_size))

        return count

    def remove_from_index(self):
        """Remove the current object from Solr by identifier using
        :meth:`index_id`"""
        # NOTE: using quotes on id to handle ids that include colons or other
        # characters that have meaning in Solr/lucene queries
        logger.debug('Deleting document from index with id %s', self.index_id())
        self.solr.update.delete_by_id([self.index_id()])




# minimal django settings required to run tests
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "test.db",
    }
}

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'parasolr',
)

# Keeping this blank avoids false detections from SECRET_KEY
# repository checkers.
SECRET_KEY = ';V1K4dLu@!U7W4sh>I+&kx:k$h!Vs%y:L-2*)s`@p?dCs:Vg:m'



# Default CI test settings; imported by parasol.solr.test_solr as testsettings
# from top level project folder
SOLR_CONNECTIONS = {
    'default': {
        # default config for testing pytest plugin
        'URL': 'http://localhost:8983/solr/',
        'COLLECTION': 'myplugin',
        'TEST': {
            'URL': 'http://localhost:8983/solr/',
            'COLLECTION': 'parasol_test',
            # aggressive commitWithin for test only
            'COMMITWITHIN': 750,
            'CONFIGSET': 'basic_configs'
        }
    }
}



import json
from typing import Any, Optional, Union
from urllib.parse import urljoin

import requests


from parasolr.solr.client import ClientBase


class Update(ClientBase):
    """API client for Solr update functionality.

    Args:
        solr_url: Base url for Solr.
        handler: Handler for Update API.
        session: A python-requests :class:`requests.Session`.
    """
    def __init__(self, solr_url: str, collection: str,
                 handler: str, commitWithin: int,
                 session: Optional[requests.Session]=None) -> None:
        # Go ahead and create a session if one is not passed in
        super().__init__(session=session)
        self.url = self.build_url(solr_url, collection, handler)
        self.headers = {'Content-Type': 'application/json'}
        self.params = {'commitWithin': commitWithin}

    def index(self, docs: list, commit: bool=False,
              commitWithin: Optional[int]=None) -> None:
        """Index a document or documents, by default with a soft commit.

        Args:
            docs (list): list of :class:`dict` objects to index.
            commit (bool, optional): Whether or not to make a hard commit to the index.
            commitWithin (int, optional): Override default commitWithin for soft commits.
        """
        params = self.params.copy()
        if commitWithin:
            params['commitWithin'] = commitWithin
        # perform a hard commit, so remove commitWithin as superfluous
        # and set params.
        if commit:
            del params['commitWithin']
            params['commit'] = True
        url = urljoin('%s/' % self.url, 'json/docs')
        self.make_request(
            'post',
            url,
            data=docs,
            params=self.params,
            headers=self.headers
        )

    def _delete(self, del_obj: Union[dict, list]) -> None:
        """Private method to pass a delete object to the update handler.

        Args:
            del_obj: Object to be serialized into valid JSON for Solr delete.
        """
        data = {'delete': del_obj}
        self.make_request(
            'post',
            self.url,
            data=data,
            params=self.params,
            headers=self.headers
        )

    def delete_by_id(self, id_list: list) -> None:
        """Delete documents by id field.

        Args:
            id_list: A list of ids.
        """
        self._delete(id_list)

    def delete_by_query(self, query: str) -> None:
        """Delete documents by an arbitrary search query.

        Args:
            query: Any valid Solr query.
        """
        self._delete({'query': query})



from unittest.mock import Mock, patch

import pytest

try:
    from django.core.exceptions import ImproperlyConfigured
    from django.test import override_settings

    from parasolr.django import SolrClient, SolrQuerySet

except ImportError:
    pass

from parasolr.tests.utils import skipif_no_django, skipif_django


@skipif_no_django
def test_django_solrclient():

    # check error handling

    # no config
    with override_settings(SOLR_CONNECTIONS=None):
        with pytest.raises(ImproperlyConfigured) as err:
            SolrClient()
        assert 'requires SOLR_CONNECTIONS in settings' in str(err)

    # config but no default
    with override_settings(SOLR_CONNECTIONS={'foo': 'bar'}):
        with pytest.raises(ImproperlyConfigured) as err:
            SolrClient()
        assert 'No "default" section in SOLR_CONNECTIONS configuration' in str(err)

    # default config but no URL
    with override_settings(SOLR_CONNECTIONS={'default': {'foo': 'bar'}}):
        with pytest.raises(ImproperlyConfigured) as err:
            SolrClient()
        assert 'No URL in default SOLR_CONNECTIONS configuration' in str(err)

    # url but no collection
    config = {'URL': 'http://my.solr.com:8943/solr'}
    with override_settings(SOLR_CONNECTIONS={'default': config}):
        solr = SolrClient()
        assert solr.solr_url == config['URL']
        assert solr.collection == ''

    # url and collection
    config['COLLECTION'] = 'mycore'
    with override_settings(SOLR_CONNECTIONS={'default': config}):
        solr = SolrClient()
        assert solr.solr_url == config['URL']
        assert solr.collection == config['COLLECTION']

    # commit within option
    config['COMMITWITHIN'] = 750
    with override_settings(SOLR_CONNECTIONS={'default': config}):
        solr = SolrClient()
        assert solr.commitWithin == 750

        # but passed in value takes precedence
        solr = SolrClient(commitWithin=7339)
        assert solr.commitWithin == 7339


@skipif_django
def test_no_django_solrclient():
    # should not be defined when django is not installed
    with pytest.raises(ImportError):
        from parasolr.solr.django import SolrClient


@skipif_no_django
@patch('parasolr.django.SolrClient')
def test_django_solrqueryset(mocksolrclient):
    # auto-initialize solr connection if not specified
    sqs = SolrQuerySet()
    mocksolrclient.assert_called_with()
    assert sqs.solr == mocksolrclient.return_value
    mocksolrclient.reset_mock()

    # use solr client if passed in
    mymocksolr = Mock(spec=SolrClient)
    sqs = SolrQuerySet(solr=mymocksolr)
    assert sqs.solr == mymocksolr
    mocksolrclient.assert_not_called()



"""
**solr_schema** is a custom manage command to update the configured
schema definition for the configured Solr instance.  Reports on the
number of fields that are added or updated, and any that are out of
date and were removed.

Example usage::

    python manage.py solr_schema

"""

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
import requests

from parasolr.django import SolrClient
from parasolr.schema import SolrSchema


class Command(BaseCommand):
    """Configure Solr schema fields and field types."""
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument(
            '--noinput',
            action='store_true',
            help='Do NOT prompt for user input'
        )

    def handle(self, *args, **kwargs):
        """Load Django solr client and project schema configuration
        and update schema field types and fields."""

        solr = SolrClient()
        noinput = kwargs.get('noinput', False)

        # check Solr connection and core exists
        try:
            core_exists = solr.core_admin.ping(solr.collection)
        except requests.exceptions.ConnectionError:
            raise CommandError('Error connecting to Solr. ' +
                               'Check your configuration and make sure Solr is running.')

        # if core does not exist, create it
        if not core_exists:
            # in no input mode, automatically create the core
            if noinput:
                create = True
            # otherwise, prompt the user to confirm
            else:
                create = input('Solr core %s does not exist. Create it? (y/n) ' %
                               solr.collection).lower() == 'y'
            if create:
                # The error handling for ensuring there's a configuration
                # has already happened, so just get default
                default_solr = settings.SOLR_CONNECTIONS['default']
                config_set = default_solr.get('CONFIGSET', 'basic_configs')
                solr.core_admin.create(solr.collection, configSet=config_set)
            else:
                # if core was not created, bail out
                return

        # find the schema configuration; error if not found or too many
        try:
            schema_config = SolrSchema.get_configuration()
        except Exception as err:
            raise CommandError(err)

        # -- configure field types
        results = schema_config.configure_fieldtypes(solr)
        # report on what was done
        self.report_changes(results, 'field type')

        # -- configure fields (includes copy fields)
        results = schema_config.configure_fields(solr)
        # report on what was done (copy fields not summarized)
        self.report_changes(results, 'field')

        # use solr core admin to trigger reload, so schema
        # changes take effect
        solr.core_admin.reload(solr.collection)

    def report_changes(self, results, label):
        """Report counts for added, replaced, or deleted items."""
        for action in ['added', 'replaced', 'deleted']:
            # if count is non-zero, report action + count + item label
            if results[action]:
                self.stdout.write(
                    '%s %d %s%s' %
                    (action.title(), results[action], label,
                     '' if results[action] == 1 else 's'))



"""
Solr schema configuration and management.

Extend :class:`SolrSchema` for your project and configure
the fields, field types, and copy fields you want defined in Solr.
Fields should be defined using :class:`SolrField` and field types
with :class:`SolrAnalyzer` and :class:`SolrFieldType`.
For example::

    from parasolr import schema

    class MySolrSchema(schema.SolrSchema):
        '''Project Solr schema configuration'''

        # field declarations
        author = schema.SolrField('text_en')
        author_exact = schema.SolrStringField()
        title = schema.SolrField('text_en')
        title_nostem = schema.SolrStringField()
        subtitle = schema.SolrField('text_en')
        collections = schema.SolrField('text_en', multivalued=True)

        #: copy fields, for facets and variant search options
        copy_fields = {
            'author': 'author_exact',
            'collections': 'collections_s',
            'title': ['title_nostem', 'title_s'],
            'subtitle': 'subtitle_s',
        }

Copy fields should be a dictionary of source and destination fields; both single
value and list are supported for destination.

If you want to define a custom field type, you can define an
analyzer for use in one or more field type declarations::

    class UnicodeTextAnalyzer(schema.SolrAnalyzer):
        '''Solr text field analyzer with unicode folding. Includes all standard
        text field analyzers (stopword filters, lower case, possessive, keyword
        marker, porter stemming) and adds ICU folding filter factory.
        '''
        tokenizer = 'solr.StandardTokenizerFactory'
        filters = [
            {"class": "solr.StopFilterFactory", "ignoreCase": True,
             "words": "lang/stopwords_en.txt"},
            {"class": "solr.LowerCaseFilterFactory"},
            {"class": "solr.EnglishPossessiveFilterFactory"},
            {"class": "solr.KeywordMarkerFilterFactory"},
            {"class": "solr.PorterStemFilterFactory"},
            {"class": "solr.ICUFoldingFilterFactory"},
        ]


    class SolrTextField(schema.SolrTypedField):
        field_type = 'text_en'

    class MySolrSchema(schema.SolrSchema):
        '''Schema configuration with custom field types'''

        text_en = schema.SolrFieldType('solr.TextField',
                                   analyzer=UnicodeFoldingTextAnalyzer)

        content = SolrTextField()


To update your configured solr core with your schema, run::

    python manage.py solr_schema

This will automatically find your :class:`SolrSchema` subclass and
apply changes.  See :mod:`~parasolr.management.commands.solr_schema`
manage command documentation for more details.

-------------------------

"""

import logging
from typing import Any, Optional

from attrdict import AttrDefault

from parasolr.solr.client import SolrClient


logger = logging.getLogger(__name__)


class SolrField:
    """A descriptor for declaring a solr field on a :class:`SolrSchema`
    instance.

    Args:
        fieldtype: The type of Solr field.
        required: Whether the field is required.
        multivalues: Whether the field is multi-valued.

    Raises:
        AttributeError: If ``__set__`` is called.
    """

    def __init__(self, fieldtype: str, required: bool=False,
                 multivalued: bool=False):
        self.type = fieldtype
        self.required = required
        self.multivalued = multivalued

    def __get__(self, obj, objtype):
        return {'type': self.type, 'required': self.required,
                'multiValued': self.multivalued}

    def __set__(self, obj, val):
        # enforce read-only descriptor
        raise AttributeError


class SolrTypedField(SolrField):
    """Base class for typed solr field descriptor. For use with your own
    field types, extend and set :attr:`field_type`.

    Args:
        *args: Arguments as passsed to :class:`SolrField`.
        **kwargs: Keyword arguments as passed to :class:`SolrField`.
    """
    field_type = None

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(self.field_type, *args, **kwargs)


class SolrStringField(SolrTypedField):
    """Solr string field."""
    field_type = 'string'


class SolrAnalyzer:
    """Class to declare a solr field analyzer with tokenizer and filters,
    for use with :class:`SolrFieldType`.
    """

    #: string name of the tokenizer to use
    tokenizer = None
    #: list of the filters to apply
    filters = None

    @classmethod
    def as_solr_config(cls):
        """ """
        return {
            'tokenizer': {
                'class': cls.tokenizer
            },
            'filters': cls.filters
        }


class SolrFieldType:
    """A descriptor for declaring and configure a solr field type on

    Args:
        field_class: The class of the SolrField
        analyzer: The name of the Solr analyzer to use on the field.

    Raises:
        AttributeError: If __set__ is called.
    """
    def __init__(self, field_class: str, analyzer: str):
        self.field_class = field_class
        self.analyzer = analyzer

    def __get__(self, obj, objtype):
        # return format neded for declaring field type
        return {
            'class': self.field_class,
            'analyzer': self.analyzer.as_solr_config()
        }

    def __set__(self, obj, val):
        # enforce read-only descriptor
        raise AttributeError


class SolrSchema:
    """Solr schema configuration."""

    #: dictionary of copy fields to be configured
    #: key is source field, value is destination field or list of fields
    copy_fields = {}

    @classmethod
    def get_configuration(cls):
        """Find a SolrSchema subclass for use as schema configuration.
        Currently only supports one schema configuration.
        """
        subclasses = cls.__subclasses__()
        if not subclasses:
            raise Exception('No Solr schema configuration found')
        elif len(subclasses) > 1:
            raise Exception('Currently only one Solr schema configuration is supported (found %d)' \
                             % len(subclasses))

        return subclasses[0]

    @classmethod
    def get_field_names(cls) -> list:
        """iterate over class attributes and return all that are instances of
        :class:`SolrField`.

        Returns:
            List of attributes that are :class:`SolrField`.
        """
        return [attr_name for attr_name, attr_type in cls.__dict__.items()
                if isinstance(attr_type, SolrField)]

    @classmethod
    def get_field_types(cls) -> list:
        """iterate over class attributes and return all that are instances of
        :class:`SolrFieldType`.

        Returns:
            List of attriubtes that are :class:`SolrFieldType`.
        """
        return [attr_name for attr_name, attr_type in cls.__dict__.items()
                if isinstance(attr_type, SolrFieldType)]

    @classmethod
    def configure_fields(cls, solr: SolrClient) -> AttrDefault:
        """Update the configured Solr instance schema to match
        the configured fields.

        Calls :meth:`configure_copy_fields` after
        new fields have been created and before old fields are removed,
        since an outdated copy field could prevent removal.

        Args:
          solr: A configured Solr instance schem.

        Returns:
            :class:`attrdict.AttrDefault` with counts for added,
            updated, and deleted fields.
        """

        current_fields = [field.name for field in solr.schema.list_fields()]
        configured_field_names = cls.get_field_names()

        # use attrdict instead of defaultdict since we have attrmap
        stats = AttrDefault(int, {})

        for field_name in configured_field_names:
            field_opts = getattr(cls, field_name)
            if field_name not in current_fields:
                logger.debug('Adding schema field %s %s', field_name, field_opts)
                solr.schema.add_field(name=field_name, **field_opts)
                stats.added += 1
            else:
                # NOTE: currently no check if field configuration has changed
                logger.debug('Replace schema field %s %s', field_name, field_opts)
                solr.schema.replace_field(name=field_name, **field_opts)
                stats.replaced += 1

        # copy fields need to be configured *after* fields are added
        # but before old fields are removed, because a copy field
        # that references an outdated field will prevent removal
        cls.configure_copy_fields(solr)

        # remove previously defined fields that are no longer current
        for field_name in current_fields:
            # don't remove special fields!
            if field_name == 'id' or field_name.startswith('_'):
                continue
            if field_name not in configured_field_names:
                stats.deleted += 1
                logger.debug('Delete schema field %s', field_name)
                solr.schema.delete_field(field_name)

        return stats

    @classmethod
    def configure_copy_fields(cls, solr: SolrClient) -> None:
        """Update configured Solr instance schema with copy fields.

        Args:
            solr: Configured Solr Schema.
        """

        # get list of currently configured copy fields
        solr_copy_fields = solr.schema.list_copy_fields()
        # create a dictionary lookup of existing copy fields from Solr
        # source field -> list of destination fields
        cp_fields = AttrDefault(list, {})
        for copyfield in solr_copy_fields:
            cp_fields[copyfield.source].append(copyfield.dest)

        # add copy fields that are not already defined
        for source, dest in cls.copy_fields.items():
            if source not in cp_fields or dest not in cp_fields[source]:
                logger.debug('Adding copy field %s %s', source, dest)
                solr.schema.add_copy_field(source, dest)

        # delete previous copy fields that are no longer wanted
        for cp_field in solr_copy_fields:
            dest = cls.copy_fields.get(cp_field.source, None)
            # check multiple conditions for copy field deletion
            delete = False
            # - source field is not in configured copy fields at all
            if cp_field.source not in cls.copy_fields:
                delete = True
            # - configured destination is a list and value is not present
            elif isinstance(dest, list):
                if cp_field.dest not in dest:
                    delete = True
            # - not a list and value does not match
            elif cp_field.dest != dest:
                delete = True

            if delete:
                logger.debug('Deleting copy field %(source)s %(dest)s', cp_field)
                solr.schema.delete_copy_field(cp_field.source, cp_field.dest)

    @classmethod
    def configure_fieldtypes(cls, solr: SolrClient) -> AttrDefault:
        """Update the configured Solr instance so the schema includes
        the configured field types, if any.

        Args:
            solr: A configured Solr instance.

        Returns:
            :class:`attrdict.AttrDefault` with counts for updated
            and added field types.
        """

        configured_field_types = cls.get_field_types()

        stats = AttrDefault(int, {})

        # if none are configured, nothing to do
        if not configured_field_types:
            return stats

        # convert list return into dictionary keyed on field type name
        current_field_types = {ftype['name']: ftype
                               for ftype in solr.schema.list_field_types()}

        for field_type in configured_field_types:
            field_type_opts = getattr(cls, field_type)
            # add name for comparison with current config
            field_type_opts['name'] = field_type
            if field_type in current_field_types:
                # if field exists [but definition has changed, ] replace it
                # NOTE: could add logic to only update when the field type
                # configuration has changed, but simple dict comparison
                # does not recognize as equal even when the config has
                # not changed
                stats.updated += 1
                logger.debug('Updating field type %s with options %s', field_type, field_type_opts)
                solr.schema.replace_field_type(**field_type_opts)

            # otherwise, create as a new field type
            else:
                stats.added += 1
                logger.debug('Adding field type %s with options %s', field_type, field_type_opts)
                solr.schema.add_field_type(**field_type_opts)

            # NOTE: currently no deletion support; would need to keep
            # a list of predefined Solr field types to check against,
            # which might change, so could be unreliable

        return stats



"""
Provides optional Django integration to automatically initialize a
:class:`parasolr.solr.SolrClient` with configurations from Django
settings.

Expected configuration format::

    SOLR_CONNECTIONS = {
        'default': {
            'URL': 'http://localhost:8983/solr/',
            'COLLECTION': 'mycore'
        }
    }

Collection can be omitted when connecting to a single-core Solr
instance.
"""

import logging
from typing import Any, Optional

try:
    import django
    from django.conf import settings
    from django.core.exceptions import ImproperlyConfigured
except ImportError:
    django = None

from parasolr.solr import client
from parasolr import query


logger = logging.getLogger(__name__)


if django:

    class SolrClient(client.SolrClient):
        """:class:`~parasolr.solr.client.SolrClient` subclass that
        automatically pulls configuration from Django settings.

        Args:
            *args: Positional arguments to be passed to :class:`parasolr.solr.client.SolrClient`.
            **kwargs: Keyword arguments to be passed to :class:`parasolr.solr.client.SolrClient`.
        """

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            solr_opts = getattr(settings, 'SOLR_CONNECTIONS', None)
            # no solr connection section at all
            if not solr_opts:
                raise ImproperlyConfigured('SolrClient requires SOLR_CONNECTIONS in settings')

            default_solr = solr_opts.get('default', None)
            # no default config
            if not default_solr:
                raise ImproperlyConfigured('No "default" section in SOLR_CONNECTIONS configuration')

            url = default_solr.get('URL', None)
            # URL is required
            if not url:
                raise ImproperlyConfigured('No URL in default SOLR_CONNECTIONS configuration')

            collection = default_solr.get('COLLECTION', '')

            # use commit within if configured
            commit_within = default_solr.get('COMMITWITHIN', None)
            # passed-in value takes precedence
            if 'commitWithin' not in kwargs and commit_within:
                kwargs['commitWithin'] = commit_within

            logger.info('Connecting to default Solr %s%s', url, collection)
            super().__init__(url, collection, *args, **kwargs)


    class SolrQuerySet(query.SolrQuerySet):
        """:class:`~parasolr.query.SolrQuerySet` subclass that
        will automatically use :class:`~parasolr.django.SolrClient` if
        no solr client is passed on.

        Args:
            Optional :class:`parasolr.solr.client.SolrClient`.
        """

        def __init__(self, solr: Optional[SolrClient] = None):
            # use passed-in solr client if there is one;
            # otherwise, initialize a django solr client
            super().__init__(solr or SolrClient())



from unittest.mock import patch, Mock, MagicMock

import pytest

try:
    from django.db.models.query import QuerySet
except ImportError:
    QuerySet = None

from parasolr.indexing import Indexable
from parasolr.tests.utils import skipif_no_django


# Define Indexable subclasses for testing
# Used for both Indexable and index manage command

class SimpleIndexable(Indexable):
    """simple indexable subclass"""
    id = 'a'

    @classmethod
    def index_item_type(cls):
        return 'simple'

    # nededed for index manage command (assumes django model)
    class objects:
        def count():
            return 5
        def all():
            return [SimpleIndexable() for i in range(5)]


class ModelIndexable(Indexable):
    """mock-model indexable subclass"""
    id = 1

    class _meta:
        verbose_name = 'model'

    # nededed for index manage command
    class objects:
        def count():
            return 1
        def all():
            return [ModelIndexable()]


@skipif_no_django
@patch('parasolr.indexing.SolrClient')
class TestIndexable:

    def test_all_indexables(self, mocksolr):
        indexables = Indexable.all_indexables()
        assert SimpleIndexable in indexables
        assert ModelIndexable in indexables

    def test_index_item_type(self, mocksolr):
        # use model verbose name by default
        assert ModelIndexable().index_item_type() == 'model'

    def test_index_id(self, mocksolr):
        assert SimpleIndexable().index_id() == 'simple.a'
        assert ModelIndexable().index_id() == 'model.1'

    def test_index_data(self, mocksolr):
        model = ModelIndexable()
        data = model.index_data()
        assert data['id'] == model.index_id()
        assert data['item_type'] == model.index_item_type()
        assert len(data) == 2

    def test_index(self, mocksolr):
        # index method on a single object instance
        model = ModelIndexable()
        model.index()
        # NOTE: because solr is stored on the class,
        # mocksolr.return_value is not the same object
        model.solr.update.index.assert_called_with([model.index_data()])

    def test_remove_from_index(self, mocksolr):
        # remove from index method on a single object instance
        model = ModelIndexable()
        model.remove_from_index()
        model.solr.update.delete_by_id.assert_called_with([model.index_id()])

    def test_index_items(self, mocksolr):
        items = [SimpleIndexable() for i in range(10)]

        indexed = Indexable.index_items(items)
        assert indexed == len(items)
        Indexable.solr.update.index \
            .assert_called_with([i.index_data() for i in items])

        # index in chunks
        Indexable.index_chunk_size = 6
        Indexable.solr.reset_mock()
        indexed = Indexable.index_items(items)
        assert indexed == len(items)
        # first chunk
        Indexable.solr.update.index \
            .assert_any_call([i.index_data() for i in items[:6]])
        # second chunk
        Indexable.solr.update.index \
            .assert_any_call([i.index_data() for i in items[6:]])

        # pass in a progressbar object
        mock_progbar = Mock()
        Indexable.index_items(items, progbar=mock_progbar)
        # progress bar update method should be called once for each chunk
        assert mock_progbar.update.call_count == 2

    def test_index_items__queryset(self, mocksolr):
        # index a queryset
        mockqueryset = MagicMock(spec=QuerySet)
        Indexable.index_items(mockqueryset)
        mockqueryset.iterator.assert_called_with()



import pytest

try:
    import django
except ImportError:
    django = None

skipif_no_django = pytest.mark.skipif(django is None,
                                      reason="requires Django")

skipif_django = pytest.mark.skipif(django,
                                   reason="requires no Django")



import os
import sys
from setuptools import find_packages, setup
from parasolr import __version__

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

REQUIREMENTS = ['requests', 'attrdict', 'progressbar2']
# NOTE: progressbar only needed for django index script; make optional?
TEST_REQUIREMENTS = ['pytest', 'pytest-cov']
DEV_REQUIREMENTS = ['sphinx', 'sphinxcontrib-napoleon',
                    'sphinx-autodoc-typehints']
# django integration is optional
DJANGO_REQUIREMENTS = ['django>=1.11', 'pytest-django']

setup(
    name='parasolr',
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    license='Apache License, Version 2.0',
    description='Lightweight python library for Solr indexing, searching' +
        ' and schema management with optional Django integration.',
    long_description=README,
    url='https://github.com/Princeton-CDH/parasolr',
    install_requires=REQUIREMENTS,
    setup_requires=['pytest-runner'],
    tests_require=TEST_REQUIREMENTS,
    extras_require={
        'test': TEST_REQUIREMENTS,
        'django': DJANGO_REQUIREMENTS,
        'dev': TEST_REQUIREMENTS + DEV_REQUIREMENTS + DJANGO_REQUIREMENTS,
    },
    author='The Center for Digital Humanities at Princeton',
    author_email='cdhdevteam@princeton.edu',
    classifiers=[
        'Environment :: Web Environment',
        'Development Status :: 2 - Pre-Alpha',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Pytest',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Database',
    ],
    entry_points={
        'pytest11': [
            'parasolr = parasolr.pytest_plugin',
        ]
    },
)



'''
**index** is a custom manage command to index content into Solr.  It
should only be run *after* your schema has been configured via
**solr_schema**.

By default, indexes _all_ indexable content.

You can optionally index specific items by type or by index id.  Default
index types are generated based on model verbose names.

A progress bar will be displayed by default if there are more than 5
items to process.  This can be suppressed via script options.

You may optionally request the index or part of the index to be cleared
before indexing, for use when index data has changed sufficiently that
previous versions need to be removed.

Example usage::

    # index everything
    python manage.py index
    # index specific items
    python manage.py index person:1 person:1 location:2
    # index one kind of item only
    python manage.py index -i person
    # suppress progressbar
    python manage.py index --no-progress
    # clear everything, then index everything
    python manage.py index --clear all
    # clear and then index one kind of item
    python manage.py index --clear person --index person
    # clear everything, index nothing
    python manage.py index --clear all --index none

'''

from django.core.management.base import BaseCommand, CommandError
from django.template.defaultfilters import pluralize
import requests
import progressbar

from parasolr.django import SolrClient
from parasolr.indexing import Indexable


class Command(BaseCommand):
    """Index content in Solr"""
    help = __doc__

    solr = None

    options = {}
    #: normal verbosity level
    v_normal = 1
    verbosity = v_normal

    indexables = {}

    def init_indexables(self):
        """Find all indexable models and create a dictionary
        keyed on index item type"""
        self.indexables = {model.index_item_type(): model
                           for model in Indexable.all_indexables()}

    def add_arguments(self, parser):
        self.init_indexables()
        # indexing choices: all, none, and all indexable model names
        choices = ['all'] + list(self.indexables.keys())
        # allow indexing none so you can clear without indexing
        index_choices = choices + ['none']

        parser.add_argument(
            'index_ids', nargs='*',
            help='List of specific items to index (optional)')
        parser.add_argument(
            '-i', '--index', choices=index_choices, default='all',
            help='Index all items or one content type (by default indexes all)')
        parser.add_argument(
            '--no-progress', action='store_true',
            help='Do not display progress bar to track the status of the reindex.')
        parser.add_argument(
            '-c', '--clear', choices=choices, required=False,
            help='Clear some or all indexed data before reindexing')

    def handle(self, *args, **kwargs):
        self.solr = SolrClient()
        self.verbosity = kwargs.get('verbosity', self.v_normal)
        self.options = kwargs

        # clear index if requested
        if self.options['clear']:
            self.clear(self.options['clear'])

        total_to_index = 0
        to_index = []

        # index specific items by id
        if self.options['index_ids']:
            # NOTE: could probably query more efficiently, but this is
            # for manual id entry so should never be very many at once
            for index_id in self.options['index_ids']:
                # relies on default format of index_id in indexable
                unrecognized_err = "Unrecognized index id '{}'".format(index_id)
                # error if id can not be split
                if Indexable.ID_SEPARATOR not in index_id:
                    raise CommandError(unrecognized_err)

                index_type, item_id = index_id.split(Indexable.ID_SEPARATOR)
                # error if split but index type is not found
                if index_type not in self.indexables:
                    raise CommandError(unrecognized_err)
                to_index.append(self.indexables[index_type].objects .get(pk=item_id))
            total_to_index = len(to_index)

        else:
            # calculate total to index across all indexables for current mode
            for name, model in self.indexables.items():
                if self.options['index'] in [name, 'all']:
                    total_to_index += model.objects.count()

        # initialize progressbar if requested and indexing more than 5 items
        progbar = None
        if not self.options['no_progress'] and total_to_index > 5:
            progbar = progressbar.ProgressBar(redirect_stdout=True,
                                              max_value=total_to_index)
        count = 0

        # index items requested
        if to_index:
            # list of objects already gathered
            count += self.index(to_index, progbar=progbar)

        else:
            # iterate over indexables by type and index if requested
            for name, model in self.indexables.items():
                if self.options['index'] in [name, 'all']:
                    # index in chunks and update progress bar
                    count += self.index(model.objects.all(), progbar=progbar)

        if progbar:
            progbar.finish()

        # commit all the indexed changes
        self.solr.update.index([], commit=True)

        # report total items indexed
        if self.verbosity >= self.v_normal:
            # using format for comma-separated numbers
            self.stdout.write('Indexed {:,} item{}'.format(
                count, pluralize(count)))

    def index(self, index_data, progbar=None):
        """Index an iterable into the configured solr"""
        try:
            # index in chunks and update progress bar if there is one
            return Indexable.index_items(index_data, progbar=progbar)
        except requests.exceptions.ConnectionError as err:
            # bail out if we error connecting to Solr
            raise CommandError(err)

    def clear(self, mode):
        '''Remove items from the Solr index.  Mode should be 'all" or
        or an item type for a configured indexable.'''
        if mode == 'all':
            del_query = '*:*'
        else:
            # construct query based on item type
            del_query = 'item_type:%s' % mode

        if self.verbosity >= self.v_normal:
            # pluralize indexable names but not all
            label = 'everything' if mode == 'all' else '%s' % mode
            self.stdout.write('Clearing %s from the index' % label)

        # return value doesn't tell us anything useful, so nothing
        # to return here
        self.solr.update.delete_by_query(del_query)



from typing import Any, Optional
from urllib.parse import urljoin

from attrdict import AttrDict
import requests

from parasolr.solr.base import ClientBase


class CoreAdmin(ClientBase):
    """API client for Solr core admin.

    Args:
        solr_url: Base url for Solr.
        handler: Handler for CoreAdmin API
        session: A python-requests :class:`requests.Session`
    """
    def __init__(self, solr_url: str, handler: str,
                 session: Optional[requests.Session]=None) -> None:
        super().__init__(session=session)
        self.solr_url = solr_url
        self.url = urljoin('%s/' % solr_url, handler)

    def create(self, name: str, **kwargs: Any) -> None:
        """Create a new core and register it.

        Args:
          name: Name of core to create.
          **kwargs: Any valid parameter for core creation in Solr.
        """
        params = {'name': name, 'action': 'CREATE'}
        params.update(kwargs)
        self.make_request('get', self.url, params=params)

    def unload(self, core: str, **kwargs: Any) -> None:
        """Unload a core, without defaults to remove data dir or index.

        Args:
          core: Name of core to unload.
          **kwargs: Any valid parameter for core unload in Solr.
        """
        params = {'core': core, 'action': 'UNLOAD'}
        params.update(kwargs)
        self.make_request('get', self.url, params=params)

    def reload(self, core: str) -> None:
        """Reload a Solr Core.

        Args:
          core: Name of core to reload.
        """
        params = {'core': core, 'action': 'RELOAD'}
        return self.make_request('get', self.url, params=params)

    def status(self, core: str=None) -> Optional[AttrDict]:
        """Get the status of all cores or one core.

        Args:
          core: Name of core to get status.
        """
        params = {}
        if core:
            params['core'] = core
        response = self.make_request('get', self.url, params=params)
        if response:
            return AttrDict(
                initFailures=response.initFailures,
                status=response.status
            )

    def ping(self, core: str) -> bool:
        """Ping a core to check status.

        Args:
          core: Name of core to ping.

        Returns:
          True if core status is OK, otherwise False.

        """
        ping_url = '/'.join([self.solr_url.rstrip('/'), core, 'admin', 'ping'])
        # ping returns 404 if core does not exist, but that's ok here
        allowed_responses = [requests.codes.ok, requests.codes.not_found]
        response = self.make_request('get', ping_url,
                                     allowed_responses=allowed_responses)
        # return True if response is valid and status is OK
        return response and response.status == 'OK'



import json
import logging
import time
from typing import Any, Optional
from urllib.parse import urljoin

import requests
from attrdict import AttrDict


logger = logging.getLogger(__name__)


class SolrClientException(Exception):
    """Base class for all exceptions in this module"""
    pass


class CoreExists(SolrClientException):
    """Raised when a Solr core exists and it should not."""
    pass


class ImproperConfiguration(SolrClientException):
    """Raised when a required setting is not present or is an invalid value."""
    pass


class ClientBase:
    """Base object with common communication methods for talking to Solr API.

    Args:
        session: A python-requests :class:`requests.Session`.
    """

    def __init__(self, session: requests.Session=None):
        if session is None:
            self.session = requests.Session()
        else:
            self.session = session


    def build_url(self, solr_url: str, collection: str, handler: str) -> str:
        """Return a url to a handler based on core and base url.

        Args:
            solr_url: Base url for Solr.
            collection: Collection or core name.
            handler: Handler URL for construction.

        Returns:
            A full-qualified URL.
        """
        # Two step proecess to avoid quirks in urljoin behavior
        # First, join the collection/core with a slashes appended
        # just in case so # it doesn't ovewrite the base url
        # (extras cause no issues)
        collection = urljoin('%s/' % solr_url, '%s/' % collection)
        # then return the core joined with handler -- without slash per
        # Solr API docs.
        return urljoin(collection, handler)

    def make_request(self, meth: str, url: str, headers: Optional[dict]=None,
                     params: Optional[dict]=None, data: Optional[dict]=None,
                     wrap: bool=True, allowed_responses: Optional[list]=None,
                     **kwargs: Any) -> Optional[AttrDict]:
        """Make an HTTP request to Solr. May optionally specify a list of
        allowed HTTP status codes for this request. Responses will be
        logged as errors if they are not in the list, but only responses
        with 200 OK status will be loaded as JSON.

        Args:
            meth: HTTP method to use.
            url: URL to make request to.
            headers: HTTP headers.
            params: Params to use as form-fields or query-string params.
            data: Data for a POST request.
            allowed_responses: HTTP status codes that are allowed for this
                request; if not set, defaults to 200 OK
            **kwargs: Any other kwargs for the request.
        """

        if params is None:
            params = dict()
            # always add wt=json for JSON api

        # copy user-supplied params for inclusion in debug logging
        user_params = params.copy()
        params['wt'] = 'json'
        # convert True and False to their appropriate string values for
        # query string to Solr
        for key, value in params.items():
            if value is True:
                params[key] = 'true'
            if value is False:
                params[key] = 'false'
        start = time.time()
        response = self.session.request(
            meth,
            url,
            params=params,
            headers=headers,
            data=json.dumps(data),
            **kwargs)
        # log the url call and speed regardless
        logger.debug(
            '%s %s => %d: %f sec%s',
            meth.upper(),
            url,
            response.status_code,
            time.time() - start,
            '\n%s' % user_params if user_params else '',
        )

        if allowed_responses is None:
            allowed_responses = [requests.codes.ok]

        if response.status_code not in allowed_responses:
            # Add the content of the response on the off chance
            # it's helpful
            logger.error(
                '%s %s=> err: %s',
                meth.upper(),
                url,
                response.content,
            )
            # return None for failure
            return

        # do further error checking on the response because Solr
        # may return 200 but pass along its own error codes and information

        # if response was allowed but not a 200, just
        # return response instead of attempting to load as json
        if response.status_code != requests.codes.ok:
            return response

        output = AttrDict(response.json())
        if 'responseHeader' in output \
                and output.responseHeader.status != 0:
            logger.error(
                '%s %s => %d: %s',
                meth.upper(),
                url,
                output.responseHeader.status,
                output.responseHeader.errors
            )
            # return None for failure
            return

        if wrap:
            return output

        return response.json()
